# -*- coding: utf-8 -*-
"""transmission control protocol

`jspcap.protocols.transport.tcp` contains `TCP` only,
which implements extractor for Transmission Control
Protocol (TCP), whose structure is described as below.

 0                   1                   2                   3
 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|          Source Port          |       Destination Port        |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                        Sequence Number                        |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                    Acknowledgment Number                      |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|  Data |           |U|A|P|R|S|F|                               |
| Offset| Reserved  |R|C|S|S|Y|I|            Window             |
|       |           |G|K|H|T|N|N|                               |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|           Checksum            |         Urgent Pointer        |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                    Options                    |    Padding    |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                             data                              |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

"""
import struct

from jspcap.corekit.infoclass import Info
from jspcap.protocols.transport.transport import Transport
from jspcap.utilities.decorators import seekset


__all__ = ['TCP']


"""TCP Option Utility Table

T | F
    bool, short of True / False

nm_len | op_len
    function, length of data bytes

chksum_opt
    dict, checksum algorithm

mptcp_opt
    dict, Multipath TCP option subtype dict

TCP_OPT
    dict, TCP option dict.
    Value is a tuple which contains:
        |--> bool, if length greater than 1
        |       |--> T - True
        |       |--> F - False
        |--> str, description string, also attribute name
        |--> (optional) function, length of data bytes
        |       |--> nm_len - default length calculation
        |       |--> op_len - optional length calculation (indicates in comment)
        |--> (optional) int, process that data bytes need (when length greater than 2)
                |--> 0: do nothing
                |--> 1: unpack according to size
                |--> 2: unpack TSopt then add to dict
                |--> 3: unpack POC-SP then add to dict
                |--> 4: unpack ACopt then fetch algorithm
                |           |--> TCP checksum
                |           |--> 8-bit Fletcher's algorithm
                |           |--> 16-bit Fletcher's algorithm
                |           |--> Redundant Checksum Avoidance
                |--> 5: unpack QSopt then add to dict
                |--> 6: unpack UTopt then add to dict
                |--> 7: unpack TCP-AO then add tot dict
                |--> 8: unpack MPTCP then add to dict
                            |--> extract subtype MP_CAPABLE
                            |--> extract subtype MP_JOIN (SYN | SYN/ACK | ACK)
                            |--> extract subtype DSS
                            |--> extract subtype ADD_ADDR
                            |--> extract subtype REMOVE_ADDR
                            |--> extract subtype MP_PRIO
                            |--> extract subtype MP_FAIL
                            |--> extract subtype MP_FASTCLOSE

"""

T = True
F = False

nm_len = lambda n: n - 2
op_len = lambda n: n * 8

chksum_opt = {  # [RFC 1146]
    0:  'TCP checksum',
    1:  "8-bit Fletcher's algorithm",
    2:  "16-bit Fletcher's algorithm",
    3:  'Redundant Checksum Avoidance',
}

mptcp_opt = {   # [RFC 6824]
    0:  lambda self, bits, size: self._read_mptcp_capable(bits, size),      # MP_CAPABLE
    1:  lambda self, bits, size: self._read_mptcp_join(bits, size),         # MP_JOIN
    2:  lambda self, bits, size: self._read_mptcp_dss(bits, size),          # DSS
    3:  lambda self, bits, size: self._read_mptcp_add(bits, size),          # ADD_ADDR
    4:  lambda self, bits, size: self._read_mptcp_remove(bits, size),       # REMOVE_ADDR
    5:  lambda self, bits, size: self._read_mptcp_prio(bits, size),         # MP_PRIO
    6:  lambda self, bits, size: self._read_mptcp_fail(bits, size),         # MP_FAIL
    7:  lambda self, bits, size: self._read_mptcp_fastclose(bits, size),    # MP_FASTCLOSE
}

process_opt = {
    0:  lambda self, size, kind: self._read_mode_donone(size, kind),    # do nothing
    1:  lambda self, size, kind: self._read_mode_unpack(size, kind),    # unpack according to size
    2:  lambda self, size, kind: self._read_mode_tsopt(size, kind),     # Timestamps
    3:  lambda self, size, kind: self._read_mode_pocsp(size, kind),     # POC Service Profile
    4:  lambda self, size, kind: self._read_mode_acopt(size, kind),     # Alternate Checksum Request
    5:  lambda self, size, kind: self._read_mode_qsopt(size, kind),     # Quick-Start Response
    6:  lambda self, size, kind: self._read_mode_utopt(size, kind),     # User Timeout Option
    7:  lambda self, size, kind: self._read_mode_tcpao(size, kind),     # TCP Authentication Option
    8:  lambda self, size, kind: self._read_mode_mptcp(size, kind),     # Multipath TCP
}

TCP_OPT = {                         #   kind  length  type  process  comment            name
    0:  (F, 'eool'),                #     0      -      -      -                [RFC 793] End of Option List
    1:  (F, 'nop'),                 #     1      -      -      -                [RFC 793] No-Operation
    2:  (T, 'mss', nm_len, 1),      #     2      4      H      1                [RFC 793] Maximum Segment Size
    3:  (T, 'ws', nm_len, 1),       #     3      3      B      1                [RFC 7323] Window Scale
    4:  (T, 'sackpmt', nm_len),     #     4      2      ?      -       True     [RFC 2018] SACK Permitted
    5:  (T, 'sack', op_len, 0),     #     5      N      P      0      2+8*N     [RFC 2018] SACK
    6:  (T, 'echo', nm_len, 0),     #     6      6      P      0                [RFC 1072][RFC 6247] Echo
    7:  (T, 'echore', nm_len, 0),   #     7      6      P      0                [RFC 1072][RFC 6247] Echo Reply
    8:  (T, 'ts', nm_len, 2),       #     8     10     II      2                [RFC 7323] Timestamps
    9:  (T, 'poc', nm_len),         #     9      2      ?      -       True     [RFC 1693][RFC 6247] POC Permitted
   10:  (T, 'pocsp', nm_len, 3),    #    10      3    ??P      3                [RFC 1693][RFC 6247] POC-Serv Profile
   11:  (T, 'cc', nm_len, 0),       #    11      6      P      0                [RFC 1693][RFC 6247] Connection Count
   12:  (T, 'ccnew', nm_len, 0),    #    12      6      P      0                [RFC 1693][RFC 6247] CC.NEW
   13:  (T, 'ccecho', nm_len, 0),   #    13      6      P      0                [RFC 1693][RFC 6247] CC.ECHO
   14:  (T, 'chkreq', nm_len, 4),   #    14      3      B      4                [RFC 1146][RFC 6247] Alt-Chksum Request
   15:  (T, 'chksum', nm_len, 0),   #    15      N      P      0                [RFC 1146][RFC 6247] Alt-Chksum Data
   19:  (T, 'sig', nm_len, 0),      #    19     18      P      0                [RFC 2385] MD5 Signature Option
   27:  (T, 'qs', nm_len, 5),       #    27      8      P      5                [RFC 4782] Quick-Start Response
   28:  (T, 'timeout', nm_len, 6),  #    28      4      P      6                [RFC 5482] User Timeout Option
   29:  (T, 'ao', nm_len, 7),       #    29      N      P      7                [RFC 5925] TCP Authentication Option
   30:  (T, 'mp', nm_len, 8),       #    30      N      P      8                [RFC 6824] Multipath TCP
   34:  (T, 'fastopen', nm_len, 0), #    34      N      P      0                [RFC 7413] Fast Open
}


class TCP(Transport):
    """This class implements Transmission Control Protocol.

    Properties:
        * name -- str, name of corresponding procotol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding procotol
        * layer -- str, `Transport`
        * length -- int, header length of corresponding protocol
        * protocol -- str, name of next layer protocol
        * protochain -- ProtoChain, protocol chain of current instance
        * src -- int, source port
        * dst -- int, destination port

    Methods:
        * read_tcp -- read Transmission Control Protocol (TCP)

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _decode_next_layer -- decode next layer protocol type
        * _import_next_layer -- import next layer protocol extractor
        * _read_tcp_options -- read TCP option list
        * _read_mode_donone -- read options request no process
        * _read_mode_unpack -- read options request unpack process
        * _read_mode_tsopt -- read Timestamps option
        * _read_mode_pocsp -- read Partial Order Connection Service Profile Option
        * _read_mode_acopt -- read Alternate Checksum Request Option
        * _read_mode_qsopt -- read Quick-Start Response Option
        * _read_mode_utopt -- read User Timeout Option
        * _read_mode_tcpao -- read Authentication Option
        * _read_mode_mptcp -- read Multipath TCP Option
        * _read_mptcp_capable -- read Multipath Capable Option
        * _read_mptcp_join -- read Join Connection Option
        * _read_join_syn -- read Join Connection Option for Initial SYN
        * _read_join_synack -- read Join Connection Option for Responding SYN/ACK
        * _read_join_ack -- read Join Connection Option for Third ACK
        * _read_mptcp_dss -- read Data Sequence Signal (Data ACK and Data Sequence Mapping) Option
        * _read_mptcp_add -- read Add Address Option
        * _read_addr_ipv4 -- read IPv4 address
        * _read_addr_ipv6 -- read IPv6 address
        * _read_mptcp_remove -- read Remove Address Option
        * _read_mptcp_prio -- read Change Subflow Priority Option
        * _read_mptcp_fail -- read Fallback Option
        * _read_mptcp_fastclose -- read Fast Close Option

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def name(self):
        """Name of current protocol."""
        return 'Transmission Control Protocol'

    @property
    def length(self):
        """Header length of current protocol."""
        return self._info.hdr_len

    @property
    def src(self):
        """Source port."""
        return self._info.srcport

    @property
    def dst(self):
        """Destination port."""
        return self._info.dstport

    ##########################################################################
    # Methods.
    ##########################################################################

    def read_tcp(self, length):
        """Read Transmission Control Protocol (TCP).

        Structure of TCP header [RFC 793]:

             0                   1                   2                   3
             0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |          Source Port          |       Destination Port        |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |                        Sequence Number                        |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |                    Acknowledgment Number                      |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |  Data |           |U|A|P|R|S|F|                               |
            | Offset| Reserved  |R|C|S|S|Y|I|            Window             |
            |       |           |G|K|H|T|N|N|                               |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |           Checksum            |         Urgent Pointer        |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |                    Options                    |    Padding    |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |                             data                              |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+


            Octets      Bits        Name                    Discription
              0           0     tcp.srcport             Source Port
              2          16     tcp.dstport             Destination Port
              4          32     tcp.seq                 Sequence Number
              8          64     tcp.ack                 Acknowledgment Number (if ACK set)
              12         96     tcp.hdr_len             Data Offset
              12        100     -                       Reserved (must be zero)
              12        103     tcp.flags.ns            ECN Concealment Protection (NS)
              13        104     tcp.flags.cwr           Congestion Window Reduced (CWR)
              13        105     tcp.flags.ecn           ECN-Echo (ECE)
              13        106     tcp.flags.urg           Urgent (URG)
              13        107     tcp.flags.ack           Acknowledgment (ACK)
              13        108     tcp.flags.push          Push Function (PSH)
              13        109     tcp.flags.reset         Reset Connection (RST)
              13        110     tcp.flags.syn           Synchronize Sequence Numbers (SYN)
              13        111     tcp.flags.fin           Last Packet from Sender (FIN)
              14        112     tcp.window_size         Size of Receive Window
              16        128     tcp.checksum            Checksum
              18        144     tcp.urgent_pointer      Urgent Pointer (if URG set)
              20        160     tcp.opt                 TCP Options (if data offset > 5)

        """
        if length is None:
            length = len(self)

        _srcp = self._read_unpack(2)
        _dstp = self._read_unpack(2)
        _seqn = self._read_unpack(4)
        _ackn = self._read_unpack(4)
        _lenf = self._read_binary(1)
        _flag = self._read_binary(1)
        _wins = self._read_unpack(2)
        _csum = self._read_fileng(2)
        _urgp = self._read_unpack(2)

        tcp = dict(
            srcport = _srcp,
            dstport = _dstp,
            seq = _seqn,
            ack = _ackn,
            hdr_len = int(_lenf[:4], base=2) * 4,
            flags = dict(
                ns = True if int(_lenf[7]) else False,
                cwr = True if int(_flag[0]) else False,
                ecn = True if int(_flag[1]) else False,
                urg = True if int(_flag[2]) else False,
                ack = True if int(_flag[3]) else False,
                push = True if int(_flag[4]) else False,
                reset = True if int(_flag[5]) else False,
                syn = True if int(_flag[6]) else False,
                fin = True if int(_flag[7]) else False,
            ),
            window_size = _wins,
            checksum = _csum,
            urgent_pointer = _urgp,
        )

        _hlen = tcp['hdr_len']
        _optl = _hlen - 20
        if _optl:
            options = self._read_tcp_options(_optl)
            tcp['opt'] = options[0]     # tuple of option acronyms
            tcp.update(options[1])      # merge option info to buffer

        length -= _hlen
        tcp['packet'] = self._read_packet(header=_hlen, payload=length)

        return self._decode_next_layer(tcp, None, length)

    ##########################################################################
    # Data models.
    ##########################################################################

    def __init__(self, _file, length=None, **kwargs):
        self._file = _file
        self._info = Info(self.read_tcp(length))

    def __length_hint__(self):
        return 20

    ##########################################################################
    # Utilities.
    ##########################################################################

    def _read_tcp_options(self, size):
        """Read TCP option list.

        Positional arguments:
            * size -- int, length of option list

        Returns:
            * dict -- extracted TCP option

        """
        counter = 0         # length of read option list
        optkind = tuple()   # option kind list
        options = dict()    # dict of option data

        while counter < size:
            # get option kind
            kind = self._read_unpack(1)

            # fetch corresponding option tuple
            opts = TCP_OPT.get(kind)
            if opts is None:
                len_ = size - counter
                counter = size
                options['Unknown'] = self._read_fileng(len_)
                break

            # extract option
            dscp = opts[1]
            if opts[0]:
                len_ = self._read_unpack(1)
                byte = opts[2](len_)
                if byte:    # check option process mode
                    data = process_opt[opts[3]](self, byte, kind)
                else:       # permission options (length is 2)
                    data = dict(
                        kind = kind,    # option kind
                        length = 2,     # option length
                        flag = True,    # permission flag
                    )
            else:           # 1-bytes options
                len_ = 1
                data = dict(
                    kind = kind,    # option kind
                    length = 1,     # option length
                )

            # record option data
            counter += len_
            if dscp in optkind:
                if isinstance(options[dscp], tuple):
                    options[dscp] += (Info(data),)
                else:
                    options[dscp] = (Info(options[dscp]), Info(data))
            else:
                optkind += (dscp,)
                options[dscp] = data

            # break when eol triggered
            if not kind:    break

        # get padding
        if counter < size:
            len_ = size - counter
            options['padding'] = self._read_fileng(len_)

        return optkind, options

    def _read_mode_donone(self, size, kind):
        """Read options request no process.

        Positional arguemnts:
            * size - int, length of option
            * kind - int, option kind value

        Returns:
            * dict -- extracted option with no operation

        Structure of TCP options:
            Octets      Bits        Name                    Discription
              0           0     tcp.opt.kind            Kind
              1           8     tcp.opt.length          Length
              2          16     tcp.opt.data            Kind-specific Data

        """
        data = dict(
            kind = kind,
            length = size,
            data = self._read_fileng(size),
        )
        return data

    def _read_mode_unpack(self, size, kind):
        """Read options request unpack process.

        Keyword arguemnts:
            size - int, length of option
            kind - int, option kind value

        Returns:
            * dict -- extracted option which unpacked

        Structure of TCP options:
            Octets      Bits        Name                    Discription
              0           0     tcp.opt.kind            Kind
              1           8     tcp.opt.length          Length
              2          16     tcp.opt.data            Kind-specific Data

        """
        data = dict(
            kind = kind,
            length = size,
            data = self._read_unpack(size),
        )
        return data

    def _read_mode_tsopt(self, size, kind):
        """Read Timestamps option.

        Positional arguemnts:
            * size - int, length of option
            * kind - int, 8 (Timestamps)

        Returns:
            * dict -- extracted Timestamps (TS) option

        Structure of TCP TSopt [RFC 7323]:
            +-------+-------+---------------------+---------------------+
            |Kind=8 |  10   |   TS Value (TSval)  |TS Echo Reply (TSecr)|
            +-------+-------+---------------------+---------------------+
                1       1              4                     4

            Octets      Bits        Name                    Discription
              0           0     tcp.ts.kind             Kind (8)
              1           8     tcp.ts.length           Length (10)
              2          16     tcp.ts.val              Timestamp Value
              6          48     tcp.ts.ecr              Timestamps Echo Reply

        """
        temp = struct.unpack('>II', self._read_fileng(size))
        data = dict(
            kind = kind,
            length = size,
            val = temp[0],
            ecr = temp[1],
        )
        return data

    def _read_mode_pocsp(self, size, kind):
        """Read Partial Order Connection Service Profile option.

        Positional arguemnts:
            * size - int, length of option
            * kind - int, 10 (POC-Serv Profile)

        Returns:
            * dict -- extracted Partial Order Connection Service Profile (POC-SP) option

        Structure of TCP POC-SP Option [RFC 1693][RFC 6247]:
                                      1 bit        1 bit    6 bits
            +----------+----------+------------+----------+--------+
            |  Kind=10 | Length=3 | Start_flag | End_flag | Filler |
            +----------+----------+------------+----------+--------+

            Octets      Bits        Name                    Discription
              0           0     tcp.pocsp.kind          Kind (10)
              1           8     tcp.pocsp.length        Length (3)
              2          16     tcp.pocsp.start         Start Flag
              2          17     tcp.pocsp.end           End Flag
              2          18     tcp.pocsp.filler        Filler

        """
        temp = self._read_binary(size)

        data = dict(
            kind = kind,
            length = size,
            start = True if int(temp[0]) else False,
            end = True if int(temp[1]) else False,
            filler = bytes(chr(int(bits[2:], base=2)), encoding='utf-8'),
        )

        return data

    def _read_mode_acopt(self, size, kind):
        """Read Alternate Checksum Request option.

        Positional arguemnts:
            size - int, length of option
            kind - int, 14 (Alt-Chksum Request)

        Returns:
            * dict -- extracted Alternate Checksum Request (CHKSUM-REQ) option

        Structure of TCP CHKSUM-REQ [RFC 1146][RFC 6247]:
            +----------+----------+----------+
            |  Kind=14 | Length=3 |  chksum  |
            +----------+----------+----------+

            Octets      Bits        Name                    Discription
              0           0     tcp.chksumreq.kind      Kind (14)
              1           8     tcp.chksumreq.length    Length (3)
              2          16     tcp.chksumreq.ac        Checksum Algorithm

        """
        temp = self._read_unpack(size)
        algo = chksum_opt.get(temp)

        data = dict(
            kind = kind,
            length = size,
            ac = algo,
        )

        return data

    def _read_mode_qsopt(self, size, kind):
        """Read Quick-Start Response option.

        Positional arguemnts:
            * size - int, length of option
            * kind - int, 27 (Quick-Start Response)

        Returns:
            * dict -- extracted Quick-Start Response (QS) option

        Structure of TCP QSopt [RFC 4782]:
             0                   1                   2                   3
             0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |     Kind      |  Length=8     | Resv. | Rate  |   TTL Diff    |
            |               |               |       |Request|               |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |                   QS Nonce                                | R |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

            Octets      Bits        Name                    Discription
              0           0     tcp.qs.kind             Kind (27)
              1           8     tcp.qs.length           Length (8)
              2          16     -                       Reserved (must be zero)
              2          20     tcp.qs.req_rate         Request Rate
              3          24     tcp.qs.ttl_diff         TTL Difference
              4          32     tcp.qs.nounce           QS Nounce
              7          62     -                       Reserved (must be zero)

        """
        rvrr = self._read_binary(1)
        ttld = self._read_unpack(1)
        noun = self._read_fileng(4)

        data = dict(
            kind = kind,
            length = size,
            req_rate = int(rvrr[4:], base=2),
            ttl_diff = ttld,
            nounce = noun[:-2],
        )

        return data

    def _read_mode_utopt(self, size, kind):
        """Read User Timeout option.

        Positional arguemnts:
            * size - int, length of option
            * kind - int, 28 (User Timeout Option)

        Returns:
            * dict -- extracted User Timeout (TIMEOUT) option

        Structure of TCP TIMEOUT [RFC 5482]:
             0                   1                   2                   3
             0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |   Kind = 28   |   Length = 4  |G|        User Timeout         |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

            Octets      Bits        Name                    Discription
              0           0     tcp.timeout.kind        Kind (28)
              1           8     tcp.timeout.length      Length (4)
              2          16     tcp.timeout.granularity Granularity
              2          17     tcp.timeout.timeout     User Timeout

        """
        temp = self._read_fileng(size)

        data = dict(
            kind = kind,
            length = size,
            granularity = 'minutes' if int(temp[0]) else 'seconds',
            timeout = bytes(chr(int(bits[0:], base=2)), encoding='utf-8'),
        )

        return data

    def _read_mode_tcpao(self, size, kind):
        """Read Authentication option.

        Positional arguemnts:
            * size - int, length of option
            * kind - int, 29 (TCP Authentication Option)

        Returns:
            * dict -- extracted Authentication (AO) option

        Structure of TCP AOopt [RFC 5925]:
            +------------+------------+------------+------------+
            |  Kind=29   |   Length   |   KeyID    | RNextKeyID |
            +------------+------------+------------+------------+
            |                     MAC           ...
            +-----------------------------------...

            ...-----------------+
            ...  MAC (con't)    |
            ...-----------------+

            Octets      Bits        Name                    Discription
              0           0     tcp.ao.kind             Kind (29)
              1           8     tcp.ao.length           Length
              2          16     tcp.ao.keyid            KeyID
              3          24     tcp.ao.rnextkeyid       RNextKeyID
              4          32     tcp.ao.mac              Message Authentication Code

        """
        key_ = self._read_unpack(1)
        rkey = self._read_unpack(1)
        mac_ = self._read_fileng(size - 2)

        data = dict(
            kind = kind,
            length = size,
            keyid = key_,
            rnextkeyid = rkey,
            mac = mac_,
        )

        return data

    def _read_mode_mptcp(self, size, kind):
        """Read Multipath TCP option.

        Positional arguemnts:
            * size - int, length of option
            * kind - int, 30 (Multipath TCP)

        Returns:
            * dict -- extracted Multipath TCP (MP-TCP) option

        Structure of MP-TCP [RFC 6824]:
                                 1                   2                   3
             0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            +---------------+---------------+-------+-----------------------+
            |     Kind      |    Length     |Subtype|                       |
            +---------------+---------------+-------+                       |
            |                     Subtype-specific data                     |
            |                       (variable length)                       |
            +---------------------------------------------------------------+

            Octets      Bits        Name                    Discription
              0           0     tcp.mp.kind             Kind (30)
              1           8     tcp.mp.length           Length
              2          16     tcp.mp.subtype          Subtype
              2          20     tcp.mp.data             Subtype-specific Data

        """
        bins = self._read_binary(1)
        subt = int(bins[:4], base=2)    # subtype number
        bits = bins[4:]                 # 4-bit data
        dlen = size - 1                 # length of remaining data

        # fetch subtype-specific data
        func = mptcp_opt.get(subt)
        if func is None:    # if subtype not exist, directly read all data
            temp = self._read_fileng(dlen)
            data = dict(
                kind = kind,
                length = size,
                subtype = 'Unknown',
                data = bytes(chr(int(bits[:4], base=2)), encoding='utf-8') + temp,
            )
        else:               # fetch corresponding subtype data dict
            data = func(self, bits, dlen, kind)
        return data

    def _read_mptcp_capable(self, bits, size, kind):
        """Read Multipath Capable option.

        Positional arguemnts:
            * bits - str, 4-bit data
            * size - int, length of option
            * kind - int, 30 (Multipath TCP)

        Returns:
            * dict -- extracted Multipath Capable (MP_CAPABLE) option

        Structure of MP_CAPABLE [RFC 6824]:
                                 1                   2                   3
             0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            +---------------+---------------+-------+-------+---------------+
            |     Kind      |    Length     |Subtype|Version|A|B|C|D|E|F|G|H|
            +---------------+---------------+-------+-------+---------------+
            |                   Option Sender's Key (64 bits)               |
            |                                                               |
            |                                                               |
            +---------------------------------------------------------------+
            |                  Option Receiver's Key (64 bits)              |
            |                     (if option Length == 20)                  |
            |                                                               |
            +---------------------------------------------------------------+

            Octets      Bits        Name                        Discription
              0           0     tcp.mp.kind                 Kind (30)
              1           8     tcp.mp.length               Length (12/20)
              2          16     tcp.mp.subtype              Subtype (0)
              2          20     tcp.mp.capable.version      Version
              3          24     tcp.mp.capable.flags.req    Checksum Require Flag (A)
              3          25     tcp.mp.capable.flags.ext    Extensibility Flag (B)
              3          26     tcp.mp.capable.flags.res    Unassigned (C-G)
              3          31     tcp.mp.capable.flags.hsa    HMAC-SHA1 Flag (H)
              4          32     tcp.mp.capable.skey         Option Sender's Key
              12         96     tcp.mp.capable.rkey         Option Receiver's Key
                                                            (if option Length == 20)

        """
        vers = int(bits, base=2)
        bins = self._read_binary(1)
        skey = self._read_fileng(8)
        rkey = self._read_fileng(8) if size == 17 else None

        data = dict(
            kind = kind,
            length = size + 1,
            subtype = 'MP_CAPABLE',
            capable = dict(
                version = vers,
                flags = dict(
                    req = True if int(bins[0]) else False,
                    ext = True if int(bins[1]) else False,
                    res = bytes(chr(int(bits[2:7], base=2)), encoding='utf-8'),
                    hsa = True if int(bins[7]) else False,
                ),
                skey = skey,
                rkey = rkey,
            ),
        )

        return data

    def _read_mptcp_join(self, bits, size, kind):
        """Read Join Connection option.

        Positional arguemnts:
            * bits - str, 4-bit data
            * size - int, length of option
            * kind - int, 30 (Multipath TCP)

        Returns:
            * dict -- extracted Join Connection (MP_JOIN) option

        Structure of MP_JOIN [RFC 6824]:
            Octets      Bits        Name                        Discription
              0           0     tcp.mp.kind                 Kind (30)
              1           8     tcp.mp.length               Length
              2          16     tcp.mp.subtype              Subtype (1)
              2          20     tcp.mp.data                 Handshake-specific Data

        """
        if self._syn and self._ack: # MP_JOIN-SYN/ACK
            return self._read_join_synack(bits, size)
        elif self._syn:             # MP_JOIN-SYN
            return self._read_join_syn(bits, size)
        elif self._ack:             # MP_JOIN-ACK
            return self._read_join_ack(bits, size)
        else:   # illegal MP_JOIN occurred
            temp = self._read_fileng(dlen)
            data = dict(
                kind = kind,
                length = size + 1,
                subtype = 'MP_JOIN-Unknown',
                data = bytes(chr(int(bits[:4], base=2)), encoding='utf-8') + temp,
            )
            return data

    def _read_join_syn(self, bits, size, kind):
        """Read Join Connection option for Initial SYN.

        Positional arguemnts:
            * bits - str, 4-bit data
            * size - int, length of option
            * kind - int, 30 (Multipath TCP)

        Returns:
            * dict -- extracted Join Connection (MP_JOIN-SYN) option for Initial SYN

        Structure of MP_JOIN-SYN [RFC 6824]:
                                 1                   2                   3
             0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            +---------------+---------------+-------+-----+-+---------------+
            |     Kind      |  Length = 12  |Subtype|     |B|   Address ID  |
            +---------------+---------------+-------+-----+-+---------------+
            |                   Receiver's Token (32 bits)                  |
            +---------------------------------------------------------------+
            |                Sender's Random Number (32 bits)               |
            +---------------------------------------------------------------+

            Octets      Bits        Name                        Discription
              0           0     tcp.mp.kind                 Kind (30)
              1           8     tcp.mp.length               Length (12)
              2          16     tcp.mp.subtype              Subtype (1|SYN)
              2          20     -                           Reserved (must be zero)
              2          23     tcp.mp.join.syn.backup      Backup Path (B)
              3          24     tcp.mp.join.syn.addrid      Address ID
              4          32     tcp.mp.join.syn.token       Receiver's Token
              8          64     tcp.mp.join.syn.randnum     Sender's Random Number

        """
        adid = self._read_unpack(1)
        rtkn = self._read_fileng(4)
        srno = self._read_unpack(4)

        data = dict(
            kind = kind,
            length = size + 1,
            subtype = 'MP_JOIN-SYN',
            join = dict(
                syn = dict(
                    backup = True if int(bits[3]) else False,
                    addrid = adid,
                    token = rtkn,
                    randnum = srno,
                ),
            ),
        )

        return data

    def _read_join_synack(self, bits, size, kind):
        """Read Join Connection option for Responding SYN/ACK.

        Positional arguemnts:
            * bits - str, 4-bit data
            * size - int, length of option
            * kind - int, 30 (Multipath TCP)

        Returns:
            * dict -- extracted Join Connection (MP_JOIN-SYN/ACK) option for Responding SYN/ACK

        Structure of MP_JOIN-SYN/ACK [RFC 6824]:
                                 1                   2                   3
             0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            +---------------+---------------+-------+-----+-+---------------+
            |     Kind      |  Length = 16  |Subtype|     |B|   Address ID  |
            +---------------+---------------+-------+-----+-+---------------+
            |                                                               |
            |                Sender's Truncated HMAC (64 bits)              |
            |                                                               |
            +---------------------------------------------------------------+
            |                Sender's Random Number (32 bits)               |
            +---------------------------------------------------------------+

            Octets      Bits        Name                        Discription
              0           0     tcp.mp.kind                 Kind (30)
              1           8     tcp.mp.length               Length (16)
              2          16     tcp.mp.subtype              Subtype (1|SYN/ACK)
              2          20     -                           Reserved (must be zero)
              2          23     tcp.mp.join.synack.backup   Backup Path (B)
              3          24     tcp.mp.join.synack.addrid   Address ID
              4          32     tcp.mp.join.synack.hmac     Sender's Truncated HMAC
              12         96     tcp.mp.join.synack.randnum  Sender's Random Number

        """
        adid = self._read_unpack(1)
        hmac = self._read_fileng(8)
        srno = self._read_unpack(4)

        data = dict(
            kind = kind,
            length = size + 1,
            subtype = 'MP_JOIN-SYN/ACK',
            join = dict(
                synack = dict(
                    backup = True if int(bits[3]) else False,
                    addrid = adid,
                    hmac = hmac,
                    randnum = srno,
                ),
            ),
        )

        return data

    def _read_join_ack(self, bits, size, kind):
        """Read Join Connection option for Third ACK.

        Positional arguemnts:
            * bits - str, 4-bit data
            * size - int, length of option
            * kind - int, 30 (Multipath TCP)

        Returns:
            * dict -- extracted Join Connection (MP_JOIN-ACK) option for Third ACK

        Structure of MP_JOIN-ACK [RFC 6824]:
                                 1                   2                   3
             0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            +---------------+---------------+-------+-----------------------+
            |     Kind      |  Length = 24  |Subtype|      (reserved)       |
            +---------------+---------------+-------+-----------------------+
            |                                                               |
            |                                                               |
            |                   Sender's HMAC (160 bits)                    |
            |                                                               |
            |                                                               |
            +---------------------------------------------------------------+

            Octets      Bits        Name                        Discription
              0           0     tcp.mp.kind                 Kind (30)
              1           8     tcp.mp.length               Length (24)
              2          16     tcp.mp.subtype              Subtype (1|ACK)
              2          20     -                           Reserved (must be zero)
              4          32     tcp.mp.join.ack.hmac        Sender's HMAC

        """
        temp = self._read_fileng(20)
        data = dict(
            kind = kind,
            length = size + 1,
            subtype = 'MP_JOIN-ACK',
            join = dict(
                ack = dict(
                    hmac = temp,
                ),
            ),
        )

        return data

    def _read_mptcp_dss(self, bits, size, kind):
        """Read Data Sequence Signal (Data ACK and Data Sequence Mapping) option.

        Positional arguemnts:
            * bits - str, 4-bit data
            * size - int, length of option
            * kind - int, 30 (Multipath TCP)

        Returns:
            * dict -- extracted Data Sequence Signal (DSS) option

        Structure of DSS [RFC 6824]:
                                 1                   2                   3
             0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            +---------------+---------------+-------+----------------------+
            |     Kind      |    Length     |Subtype| (reserved) |F|m|M|a|A|
            +---------------+---------------+-------+----------------------+
            |           Data ACK (4 or 8 octets, depending on flags)       |
            +--------------------------------------------------------------+
            |   Data sequence number (4 or 8 octets, depending on flags)   |
            +--------------------------------------------------------------+
            |              Subflow Sequence Number (4 octets)              |
            +-------------------------------+------------------------------+
            |  Data-Level Length (2 octets) |      Checksum (2 octets)     |
            +-------------------------------+------------------------------+

                                1                   2                   3
             0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            +--------------------------------------------------------------+
            |                                                              |
            |                Data Sequence Number (8 octets)               |
            |                                                              |
            +--------------------------------------------------------------+
            |              Subflow Sequence Number (4 octets)              |
            +-------------------------------+------------------------------+
            |  Data-Level Length (2 octets) |        Zeros (2 octets)      |
            +-------------------------------+------------------------------+

            Octets      Bits        Name                        Discription
              0           0     tcp.mp.kind                 Kind (30)
              1           8     tcp.mp.length               Length
              2          16     tcp.mp.subtype              Subtype (2)
              2          20     -                           Reserved (must be zero)
              3          27     tcp.mp.dss.flags.fin        DATA_FIN (F)
              3          28     tcp.mp.dss.flags.dsn_len    DSN Length (m)
              3          29     tcp.mp.dss.flags.data_pre   DSN, SSN, Data-Level Length, CHKSUM Present (M)
              3          30     tcp.mp.dss.flags.ack_len    ACK Length (a)
              3          31     tcp.mp.dss.flags.ack_pre    Data ACK Present (A)
              4          32     tcp.mp.dss.ack              Data ACK (4/8 octets)
              8-12    64-96     tcp.mp.dss.dsn              DSN (4/8 octets)
              12-20  48-160     tcp.mp.dss.ssn              Subflow Sequence Number
              16-24 128-192     tcp.mp.dss.dl_len           Data-Level Length
              18-26 144-208     tcp.mp.dss.checksum         Checksum

        """
        bits = self._read_binary(1)
        mflg = 8 if int(bits[4]) else 4
        Mflg = True if int(bits[5]) else False
        aflg = 8 if int(bits[6]) else 4
        Aflg = True if int(bits[7]) else False
        ack_ = self._read_fileng(aflg) if Aflg else None
        dsn_ = self._read_unpack(mflg) if Mflg else None
        ssn_ = self._read_unpack(4) if Mflg else None
        dll_ = self._read_unpack(2) if Mflg else None
        chk_ = self._read_fileng(2) if Mflg else None

        data = dict(
            kind = kind,
            length = size + 1,
            subtype = 'DSS',
            dss = dict(
                flags = dict(
                    fin = True if int(bits[3]) else False,
                    dsn_len = mflg,
                    data_pre = Mflg,
                    ack_len = aflg,
                    ack_pre = Aflg,
                ),
                ack = ack_,
                dsn = dsn_,
                ssn = ssn_,
                dl_len = dll_,
                checksum = chk,
            ),
        )

        return data

    def _read_mptcp_add(self, bits, size, kind):
        """Read Add Address option.

        Positional arguemnts:
            * bits - str, 4-bit data
            * size - int, length of option
            * kind - int, 30 (Multipath TCP)

        Returns:
            * dict -- extracted Add Address (ADD_ADDR) option

        Structure of ADD_ADDR [RFC 6824]:
                                 1                   2                   3
             0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            +---------------+---------------+-------+-------+---------------+
            |     Kind      |     Length    |Subtype| IPVer |  Address ID   |
            +---------------+---------------+-------+-------+---------------+
            |          Address (IPv4 - 4 octets / IPv6 - 16 octets)         |
            +-------------------------------+-------------------------------+
            |   Port (2 octets, optional)   |
            +-------------------------------+

            Octets      Bits        Name                        Discription
              0           0     tcp.mp.kind                 Kind (30)
              1           8     tcp.mp.length               Length
              2          16     tcp.mp.subtype              Subtype (3)
              2          20     tcp.mp.addaddr.ipver        IP Version
              3          24     tcp.mp.addaddr.addrid       Address ID
              4          32     tcp.mp.addaddr.addr         IP Address (4/16)
              8-20   64-160     tcp.mp.addaddr.port         Port (optional)

        """
        vers = int(bits, base=2)
        adid = self._read_unpack(1)
        ipad = self._read_addr_ipv4() if vers == 4 else self._read_addr_ipv6()
        ip_l = 4 if vers == 4 else 16
        pt_l = size - 1 - ip_l
        port = self._read_unpack(2) if pt_l else None

        data = dict(
            kind = kind,
            length = size + 1,
            subtype = 'ADD_ADDR',
            addaddr = dict(
                ipver = vers,
                addrid = adid,
                addr = ipad,
                port = port,
            ),
        )

        return data

    def _read_addr_ipv4(self):
        """Read IPv4 address."""
        byte = self._read_fileng(4)
        addr = '.'.join([str(_) for _ in byte])
        return addr

    def _read_addr_ipv6(self):
        """Read IPv6 address."""
        adlt = []       # list of IPv6 hexadecimal address
        ctr_ = collections.defaultdict(int)
                        # counter for consecutive groups of zero value
        ptr_ = 0        # start pointer of consecutive groups of zero value
        last = False    # if last hextet/group is zero value
        ommt = False    # ommitted flag, since IPv6 address can ommit to `::` only once

        for _ in range(8):
            hex_ = self._read_fileng(2).hex().lstrip('0')

            if hex_:    # if hextet is not '', directly append
                adlt.append(hex_)
                last = False
            else:       # if hextet is '', append '0'
                adlt.append('0')
                if last:    # if last hextet is '', ascend counter
                    ctr_[ptr_] += 1
                else:       # if last hextet is not '', record pointer
                    ptr_ = _
                    last = True
                    ctr_[ptr_] = 1

        ptr_ = max(ctr_, key=ctr_.get) if ctr_ else 0   # fetch start pointer with longest zero values
        end_ = ptr_ + ctr_[ptr_]                        # calculate end pointer

        if ctr_[ptr_] > 1:      # only ommit if zero values are in a consecutive group
            del adlt[ptr_:end_] # remove zero values

            if ptr_ == 0 and end_ == 8:     # insert `::` if IPv6 unspecified address (::)
                adlt.insert(ptr_, '::')
            elif ptr_ == 0 or end_ == 8:    # insert `:` if zero values are from start or at end
                adlt.insert(ptr_, ':')
            else:                           # insert '' otherwise
                adlt.insert(ptr_, '')

        addr = ':'.join(adlt)
        return addr

    def _read_mptcp_remove(self, bits, size):
        """Read Remove Address option.

        Positional arguemnts:
            * bits - str, 4-bit data
            * size - int, length of option

        Returns:
            * dict -- extracted Remove Address (REMOVE_ADDR) option

        Structure of REMOVE_ADDR [RFC 6824]:
                                 1                   2                   3
             0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            +---------------+---------------+-------+-------+---------------+
            |     Kind      |  Length = 3+n |Subtype|(resvd)|   Address ID  | ...
            +---------------+---------------+-------+-------+---------------+
                                       (followed by n-1 Address IDs, if required)

            Octets      Bits        Name                            Discription
              0           0     tcp.opt.kind                    Kind (30)
              1           8     tcp.opt.length                  Length
              2          16     tcp.opt.mp.subtype              Subtype (4)
              2          20     -                               Reserved (must be zero)
              3          24     tcp.opt.mp.removeaddr.addrid    Address ID (optional list)

        """
        adid = []
        for _ in size:
            adid.append(self._read_unpack(1))

        data = dict(
            subtype = 'REMOVE_ADDR',
            removeaddr = dict(
                addrid = adid or None,
            ),
        )

        return data

    def _read_mptcp_prio(self, bits, size):
        """Read Change Subflow Priority option.

        Positional arguemnts:
            * bits - str, 4-bit data
            * size - int, length of option

        Returns:
            * dict -- extracted Change Subflow Priority (MP_PRIO) option

        Structure of MP_PRIO [RFC 6824]:
                                  1                   2                   3
              0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            +---------------+---------------+-------+-----+-+--------------+
            |     Kind      |     Length    |Subtype|     |B| AddrID (opt) |
            +---------------+---------------+-------+-----+-+--------------+

            Octets      Bits        Name                    Discription
              0           0     tcp.opt.kind            Kind (30)
              1           8     tcp.opt.length          Length (3/4)
              2          16     tcp.opt.mp.subtype      Subtype (5)
              2          23     tcp.opt.mp.prio.backup  Backup Path (B)
              3          24     tcp.opt.mp.prio.addrid  Address ID (optional)

        """
        temp = self._read_unpack(1) if size else None
        data = dict(
            subtype = 'MP_PRIO',
            prio = dict(
                res = b'\x00' * 3,
                backup = True if int(bits[3]) else False,
                addrid = temp,
            ),
        )
        return data

    def _read_mptcp_fail(self, bits, size):
        """Read Fallback option.

        Positional arguemnts:
            * bits - str, 4-bit data
            * size - int, length of option

        Returns:
            * dict -- extracted Fallback (MP_FAIL) option

        Structure of MP_FAIL [RFC 6824]:
                                 1                   2                   3
             0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            +---------------+---------------+-------+----------------------+
            |     Kind      |   Length=12   |Subtype|      (reserved)      |
            +---------------+---------------+-------+----------------------+
            |                                                              |
            |                 Data Sequence Number (8 octets)              |
            |                                                              |
            +--------------------------------------------------------------+

            Octets      Bits        Name                    Discription
              0           0     tcp.opt.kind            Kind (30)
              1           8     tcp.opt.length          Length (12)
              2          16     tcp.opt.mp.subtype      Subtype (6)
              2          23     -                       Reserved (must be zero)
              4          32     tcp.opt.mp.fail.dsn     Data Sequence Number

        """
        resv = self._read_fileng(1)
        dsn_ = self._read_unpack(8)

        data = dict(
            subtype = 'MP_FAIL',
            fail = dict(
                dsn = dsn_,
            ),
        )

        return data

    def _read_mptcp_fastclose(self, bits, size):
        """Read Fast Close option.

        Positional arguemnts:
            * bits - str, 4-bit data
            * size - int, length of option

        Returns:
            * dict -- extracted Fast Close (MP_FASTCLOSE) option

        Structure of MP_FASTCLOSE [RFC 6824]:
                                 1                   2                   3
             0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            +---------------+---------------+-------+-----------------------+
            |     Kind      |    Length     |Subtype|      (reserved)       |
            +---------------+---------------+-------+-----------------------+
            |                      Option Receiver's Key                    |
            |                            (64 bits)                          |
            |                                                               |
            +---------------------------------------------------------------+

            Octets      Bits        Name                        Discription
              0           0     tcp.opt.kind                Kind (30)
              1           8     tcp.opt.length              Length (12)
              2          16     tcp.opt.mp.subtype          Subtype (7)
              2          23     -                           Reserved (must be zero)
              4          32     tcp.opt.mp.fastclose.rkey   Option Receiver's Key

        """
        resv = self._read_fileng(1)
        rkey = self._read_fileng(8)

        data = dict(
            subtype = 'MP_FASTCLOSE',
            fastclose = dict(
                rkey = rkey,
            ),
        )

        return data



# -*- coding: utf-8 -*-
"""index for the library

`jspcap` has defined various and numerous functions and
classes, which have different features and purposes. To
make a simple index for this library, `jspcap.all`
contains all things from `jspcap`.

"""
from jspcap.corekit import *
from jspcap.dumpkit import *
from jspcap.foundation import *
from jspcap.interface import *
from jspcap.ipsuite import *
from jspcap.protocols import *
from jspcap.reassembly import *
from jspcap.utilities import *


__all__ = [
    # jspcap.corekit
    'Info',                                                 # Info Class
    'ProtoChain',                                           # ProtoChain
    'VersionInfo',                                          # Version

    # jspcap.dumpkit
    'PCAP_Dumper',                                          # PCAP Dumper

    # jspcap.foundation
    'Extractor',                                            # Extraction
    'Analysis',                                             # Analysis
    'TraceFlow',                                            # Trace Flow

    # jspcap.interface
    'extract', 'analyse', 'reassemble', 'trace',            # Functions
    'TREE', 'JSON', 'PLIST', 'PCAP',                        # Macros

    # jspcap.ipsuite
    'IPSHeader', 'IPSFrame',                                # PCAP Headers

    # jspcap.protocols
    'LINKTYPE', 'ETHERTYPE', 'TP_PROTO',                    # Protocol Numbers
    'Header', 'Frame',                                      # PCAP Headers
    'Raw',                                                  # Raw Packet
    'ARP', 'DRARP', 'Ethernet', 'InARP', 'L2TP', 'OSPF', 'RARP', 'VLAN',
                                                            # Link Layer
    'AH', 'IP', 'IPsec', 'IPv4', 'IPv6', 'IPX',             # Internet Layer
    'HIP', 'HOPOPT', 'IPv6_Frag', 'IPv6_Opts', 'IPv6_Route', 'MH',
                                                            # IPv6 Extension Header
    'TCP', 'UDP',                                           # Transport Layer
    'HTTP',                                                 # Application Layer

    # jspcap.reassembly
    'IPv4_Reassembly', 'IPv6_Reassembly',                   # IP Reassembly
    'TCP_Reassembly',                                       # TCP Reassembly

    # jspcap.utilities
    'beholder_ng', 'seekset_ng',                            # Decorators
]



# -*- coding: utf-8 -*-
"""command line tool

`jspcap.__main__` was originally the module file of
[`jspcapy`](https://github.com/JarryShaw/jspcapy), which
is now deprecated and merged with `jspcap`.

"""
import argparse
import os
import sys
import warnings

from jspcap.foundation.extraction import Extractor
from jspcap.interface import TREE, JSON, PLIST


# version number
__version__ = '0.5.0'


def get_parser():
    parser = argparse.ArgumentParser(prog='jspcapy', description=(
        'PCAP file extractor and formatted exporter'
    ))
    parser.add_argument('-V', '--version', action='version', version=f'{__version__}')
    parser.add_argument('fin', metavar='input-file-name',
                        help=(
                            'The name of input pcap file. If ".pcap" omits, '
                            'it will be automatically appended.'
                        ))
    parser.add_argument('-o', '--output', action='store', metavar='file-name',
                        dest='fout', help=(
                            'The name of input pcap file. If format extension '
                            'omits, it will be automatically appended.'
                        ))
    parser.add_argument('-f', '--format', action='store', metavar='format',
                        dest='format', help=(
                            'Print a extraction report in the specified output '
                            'format. Available are all formats supported by '
                            'jsformat, e.g.: json, plist, and tree.'
                        ))
    parser.add_argument('-j', '--json', action='store_true', default=False,
                        help=(
                            'Display extraction report as json. This will yield '
                            '"raw" output that may be used by external tools. '
                            'This option overrides all other options.'
                        ))
    parser.add_argument('-p', '--plist', action='store_true', default=False,
                        help=(
                            'Display extraction report as macOS Property List '
                            '(plist). This will yield "raw" output that may be '
                            'used by external tools. This option overrides all '
                            'other options.'
                        ))
    parser.add_argument('-t', '--tree', action='store_true', default=False,
                        help=(
                            'Display extraction report as tree view text. This '
                            'will yield "raw" output that may be used by external '
                            'tools. This option overrides all other options.'
                        ))
    parser.add_argument('-a', '--auto-extension', action='store_true', default=False,
                        help=(
                            'If output file extension omits, append automatically.'
                        ))
    parser.add_argument('-v', '--verbose', action='store_false', default=True,
                        help=(
                            'Show more information.'
                        ))
    parser.add_argument('-F', '--files', action='store_true', default=False,
                        help=(
                            'Split each frame into different files.'
                        ))
    parser.add_argument('-E', '--engine', action='store', dest='engine', choises=[
                        'default', 'jspcap', 'dpkt', 'scapy', 'pyshark',
                        ], default='default', metavar='PKG', help=(
                            'Indicate extraction engine. Note that except '
                            'default or jspcap engine, all other engines '
                            'need support of correspoding packages.'
                        ))
    parser.add_argument('-P', '--protocol', action='store', dest='protocol',
                        default='null', metavar='PROTOCOL', help=(
                            'Indicate extraction stops after which protocol.'
                        ))
    parser.add_argument('-L', '--layer', action='store', dest='layer', choises=[
                        'None', 'Link', 'Internet', 'Transport', 'Application',
                        ], default='None', metavar='LAYER', help=(
                            'Indicate extract frames until which layer.'
                        ))
    return parser


def main():
    parser = get_parser()
    args = parser.parse_args()
    warnings.simplefilter('ignore')

    if args.format:     fmt = args.format
    elif args.json:     fmt = JSON
    elif args.plist:    fmt = PLIST
    elif args.tree:     fmt = TREE
    else:               fmt = None

    extractor = Extractor(store=False, format=fmt,
                            fin=args.fin, fout=args.fout,
                            auto=args.verbose, files=args.files,
                            layer=args.layer, protocol=args.protocol,
                            engine=args.engine, extension=args.auto_extension)

    if not args.verbose:
        print(f"🚨 Loading file '{extractor.input}'")
        for frame in extractor:
            print(f' - Frame {extractor.length:>3d}: {extractor.protocol}')
        print(f"🍺 Report file{'s' if args.files else ''} stored in '{extractor.output}'")


if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-
"""internet protocol version 4

`jspcap.protocols.internet.ipv4` contains `IPv4` only,
which implements extractor for Internet Protocol version 4
(IPv4), whose structure is described as below.

 0                   1                   2                   3
 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|Version|  IHL  |Type of Service|          Total Length         |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|         Identification        |Flags|      Fragment Offset    |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|  Time to Live |    Protocol   |         Header Checksum       |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                       Source Address                          |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                    Destination Address                        |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                    Options                    |    Padding    |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

"""
import collections
import datetime
import ipaddress

from jspcap.corekit.infoclass import Info
from jspcap.corekit.protochain import ProtoChain
from jspcap.protocols.internet.ip import IP
from jspcap.utilities.exceptions import ProtocolError


__all__ = ['IPv4']


# TOS (DS Field) Precedence
TOS_PRE = {
    '111':  'Network Control',
    '110':  'Internetwork Control',
    '101':  'CRITIC/ECP',
    '100':  'Flash Override',
    '011':  'Flash',
    '010':  'Immediate',
    '001':  'Priority',
    '000':  'Routine',
}

# TOS (DS Field) Delay
TOS_DEL = {
    '0':    'Normal Delay',
    '1':    'Low Delay',
}

# TOS (DS Field) Throughput
TOS_THR = {
    '0':    'Normal Throughput',
    '1':    'High Throughput',
}

# TOS (DS Field) Relibility
TOS_REL = {
    '0':    'Normal Relibility',
    '1':    'High Relibility',
}

# TOS ECN FIELD
TOS_ECN = {
    '00':   'Not-ECT',
    '01':   'ECT(1)',
    '10':   'ECT(0)',
    '11':   'CE',
}

# QS Functions
QS_FUNC = {
    1:  'Quick-Start Request',
    2:  'Report of Approved Rate',
}

# Classification Level Encodings
_CLASSIFICATION_LEVEL = {
    '00000001' : 'Reserved (4)',
    '00111101' : 'Top Secret',
    '01011010' : 'Secret',
    '10010110' : 'Confidential',
    '01100110' : 'Reserved (3)',
    '11001100' : 'Reserved (2)',
    '10101011' : 'Unclassified',
    '11110001' : 'Reserved (1)',
}

# Protection Authority Bit Assignments
_PROTECTION_AUTHORITY = {
    0 : 'GENSER',
    1 : 'SIOP-ESI',
    2 : 'SCI',
    3 : 'NSA',
    4 : 'DOE',
    5 : 'Unassigned',
    6 : 'Unassigned',
    7 : 'Field Termination Indicator',
}

# Router Alert Code
_ROUTER_ALERT = {
    1 : 'Router shall examine packet',
}


"""IPv4 Option Utility Table

T | F
    bool, short of True / False

opt_class
    dict, option classes

IPv4_OPT
    dict, IPv4 option dict.
    Value is a tuple which contains:
        |--> bool, if length greater than 1
        |       |--> T - True
        |       |--> F - False
        |--> str, description string, also attribute name
        |--> (optional) int, process that data bytes need (when length greater than 2)
                |--> 0: do nothing
                |--> 1: unpack according to size
                |--> 2: unpack route data options then add to dict
                |--> 3: unpack Quick-Start then add to dict
                |--> 4: unpack Time Stamp then add to dict
                |--> 5: unpack Traceroute then add to dict
                |--> 6: unpack (Extended) Security then add tot dict
                |--> 7: unpack Router Alert then add to dict

"""

T = True
F = False

opt_class = {
    0:  'control',
    1:  'reserved for future use',
    2:  'debugging and measurement',
    3:  'reserved for future use',
}

process_opt = {
    0:  lambda self, size, kind: self._read_mode_donone(size, kind),    # do nothing
    1:  lambda self, size, kind: self._read_mode_unpack(size, kind),    # unpack according to size
    2:  lambda self, size, kind: self._read_mode_route(size, kind),     # route data
    3:  lambda self, size, kind: self._read_mode_qs(size, kind),        # Quick-Start
    4:  lambda self, size, kind: self._read_mode_ts(size, kind),        # Time Stamp
    5:  lambda self, size, kind: self._read_mode_tr(size, kind),        # Traceroute
    6:  lambda self, size, kind: self._read_mode_sec(size, kind),       # (Extended) Security
    7:  lambda self, size, kind: self._read_mode_rsralt(size, kind),    # Router Alert
}

IPv4_OPT = {                        #   copy  class  number  kind  length  process          name
    0:  (F, 'eool'),                #     0     0       0      0      -       -     [RFC 791] End of Option List
    1:  (F, 'nop'),                 #     0     0       1      1      -       -     [RFC 791] No-Operation
    7:  (T, 'rr', 2),               #     0     0       7      7      N       2     [RFC 791] Record Route
   11:  (T, 'mtup', 1),             #     0     0      11     11      4       1     [RFC 1063][RFC 1191] MTU Probe
   12:  (T, 'mtur', 1),             #     0     0      12     12      4       1     [RFC 1063][RFC 1191] MTU Reply
   25:  (T, 'qs', 3),               #     0     0      25     25      8       3     [RFC 4782] Quick-Start
   68:  (T, 'ts', 4),               #     0     2       4     68      N       4     [RFC 791] Time Stamp
   82:  (T, 'tr', 5),               #     0     2      18     82      N       5     [RFC 1393][RFC 6814] Traceroute
  130:  (T, 'sec', 6),              #     1     0       2    130      N       6     [RFC 1108] Security
  131:  (T, 'lsr', 2),              #     1     0       3    131      N       2     [RFC 791] Loose Source Route
  133:  (T, 'esec', 6),             #     1     0       5    133      N       6     [RFC 1108] Extended Security
  136:  (T, 'sid', 1),              #     1     0       8    136      4       1     [RFC 791][RFC 6814] Stream ID
  137:  (T, 'ssr', 2),              #     1     0       9    137      N       2     [RFC 791] Strict Source Route
  145:  (T, 'eip', 0),              #     1     0      17    145      N       0     [RFC 1385][RFC 6814] Extended Internet Protocol
  148:  (T, 'rtralt', 7),           #     1     0      20    148      4       7     [RFC 2113] Router Alert
}


class IPv4(IP):
    """This class implements Internet Protocol version 4.

    Properties:
        * name -- str, name of corresponding procotol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding procotol
        * layer -- str, `Internet`
        * length -- int, header length of corresponding protocol
        * protocol -- str, name of next layer protocol
        * protochain -- ProtoChain, protocol chain of current instance
        * src -- str, source IP address
        * dst -- str, destination IP address

    Methods:
        * read_ipv4 -- read Internet Protocol version 4 (IPv4)

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _decode_next_layer -- decode next layer protocol type
        * _import_next_layer -- import next layer protocol extractor
        * _read_ipv4_addr -- read IPv4 address
        * _read_ipv4_options -- read IPv4 option list

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def name(self):
        """Name of corresponding procotol."""
        return 'Internet Protocol version 4'

    @property
    def length(self):
        """Header length of corresponding protocol."""
        return self._info.hdr_len

    @property
    def protocol(self):
        """Name of next layer protocol."""
        return self._info.proto

    ##########################################################################
    # Methods.
    ##########################################################################

    def read_ipv4(self, length):
        """Read Internet Protocol version 4 (IPv4).

        Structure of IPv4 header [RFC 791]:

             0                   1                   2                   3
             0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |Version|  IHL  |Type of Service|          Total Length         |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |         Identification        |Flags|      Fragment Offset    |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |  Time to Live |    Protocol   |         Header Checksum       |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |                       Source Address                          |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |                    Destination Address                        |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |                    Options                    |    Padding    |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

            Octets      Bits        Name                    Discription
              0           0     ip.version              Version (4)
              0           4     ip.hdr_len              Interal Header Length (IHL)
              1           8     ip.dsfield.dscp         Differentiated Services Code Point (DSCP)
              1          14     ip.dsfield.ecn          Explicit Congestion Notification (ECN)
              2          16     ip.len                  Total Length
              4          32     ip.id                   Identification
              6          48     -                       Reserved Bit (must be zero)
              6          49     ip.flags.df             Don't Fragment (DF)
              6          50     ip.flags.mf             More Fragments (MF)
              6          51     ip.frag_offset          Fragment Offset
              8          64     ip.ttl                  Time To Live (TTL)
              9          72     ip.proto                Protocol (Transport Layer)
              10         80     ip.checksum             Header Checksum
              12         96     ip.src                  Source IP Address
              16        128     ip.dst                  Destination IP Address
              20        160     ip.options              IP Options (if IHL > 5)

        """
        if length is None:
            length = len(self)

        _vihl = self._read_fileng(1).hex()
        _dscp = self._read_binary(1)
        _tlen = self._read_unpack(2)
        _iden = self._read_unpack(2)
        _frag = self._read_binary(2)
        _ttol = self._read_unpack(1)
        _prot = self._read_protos(1)
        _csum = self._read_fileng(2)
        _srca = self._read_ipv4_addr()
        _dsta = self._read_ipv4_addr()

        ipv4 = dict(
            version = _vihl[0],
            hdr_len = int(_vihl[1], base=16) * 4,
            dsfield = dict(
                dscp = (
                    TOS_PRE.get(_dscp[:3]),
                    TOS_DEL.get(_dscp[3]),
                    TOS_THR.get(_dscp[4]),
                    TOS_REL.get(_dscp[5]),
                ),
                ecn = TOS_ECN.get(_dscp[-2:]),
            ),
            len = _tlen,
            id = _iden,
            flags = dict(
                df = True if int(_frag[1]) else False,
                mf = True if int(_frag[2]) else False,
            ),
            frag_offset = int(_frag[3:], base=2) * 8,
            ttl = _ttol,
            proto = _prot,
            checksum = _csum,
            src = _srca,
            dst = _dsta,
        )

        _optl = ipv4['hdr_len'] - 20
        if _optl:
            options = self._read_ipv4_options(_optl)
            ipv4['opt'] = options[0]    # tuple of option acronyms
            ipv4.update(options[1])     # merge option info to buffer
            # ipv4['opt'] = self._read_fileng(_optl) or None

        hdr_len = ipv4['hdr_len']
        raw_len = ipv4['len'] - hdr_len
        ipv4['packet'] = self._read_packet(header=hdr_len, payload=raw_len)

        return self._decode_next_layer(ipv4, _prot, raw_len)

    ##########################################################################
    # Data models.
    ##########################################################################

    def __init__(self, _file, length=None, **kwargs):
        self._file = _file
        self._info = Info(self.read_ipv4(length))

    def __length_hint__(self):
        return 20

    @classmethod
    def __index__(cls):
        return cls.__name__

    ##########################################################################
    # Utilities.
    ##########################################################################

    def _read_ipv4_addr(self):
        """Read IP address."""
        # _byte = self._read_fileng(4)
        # _addr = '.'.join([str(_) for _ in _byte])
        # return _addr
        return ipaddress.ip_address(self._read_fileng(4))

    def _read_opt_type(self, kind):
        """Read option type field.

        Positional arguments:
            * kind -- int, option kind value

        Returns:
            * dict -- extracted IPv4 option

        Structure of option type field [RFC 791]:

            Octets      Bits        Name                    Discriptions
              0           0     ip.opt.type.copy        Copied Flag (0/1)
              0           1     ip.opt.type.class       Option Class (0-3)
              0           3     ip.opt.type.number      Option Number

        """
        bin_ = bin(kind)[2:].zfill(8)
        type_ = {
            'copy' : bool(int(bin_[0], base=2)),
            'class' : opt_class.get(int(bin_[1:3], base=2)),
            'number' : int(bin_[3:], base=2),
        }
        return type_

    def _read_ipv4_options(self, size=None):
        """Read IPv4 option list.

        Keyword arguments:
            * size -- int, buffer size

        """
        counter = 0         # length of read option list
        optkind = tuple()   # option kind list
        options = dict()    # dict of option data

        while counter < size:
            # get option kind
            kind = self._read_unpack(1)

            # fetch corresponding option tuple
            opts = IPv4_OPT.get(kind)
            if opts is None:
                len_ = size - counter
                counter = size
                options['Unknown'] = self._read_fileng(len_)
                break

            # extract option
            dscp = opts[1]
            if opts[0]:
                byte = self._read_unpack(1)
                if byte:    # check option process mode
                    data = process_opt[opts[2]](self, byte, kind)
                else:       # permission options (length is 2)
                    data = dict(
                        kind = kind,                        # option kind
                        type = self._read_opt_type(kind),   # option type info
                        length = 2,                         # option length
                        flag = True,                        # permission flag
                    )
            else:           # 1-bytes options
                byte = 1
                data = dict(
                    kind = kind,                        # option kind
                    type = self._read_opt_type(kind),   # option type info
                    length = 1,                         # option length
                )

            # record option data
            counter += byte
            if dscp in optkind:
                if isinstance(options[dscp], tuple):
                    options[dscp] += (Info(data),)
                else:
                    options[dscp] = (Info(options[dscp]), Info(data))
            else:
                optkind += (dscp,)
                options[dscp] = data

            # break when eol triggered
            if not kind:    break

        # get padding
        if counter < size:
            len_ = size - counter
            padding = self._read_binary(len_)
            if any((int(bit, base=2) for bit in padding)):
                raise ProtocolError(f'{self.alias}: invalid format')

        return optkind, options

    def _read_mode_donone(self, size, kind):
        """Read options request no process.

        Positional arguemnts:
            * size - int, length of option
            * kind - int, option kind value

        Returns:
            * dict -- extracted option

        Structure of TCP options:
            Octets      Bits        Name                    Discription
              0           0     ip.opt.kind             Kind
              0           0     ip.ts.type.copy         Copied Flag
              0           1     ip.ts.type.class        Option Class
              0           3     ip.ts.type.number       Option Number
              1           8     ip.opt.length           Length
              2          16     ip.opt.data             Kind-specific Data

        """
        if size < 3:
            raise ProtocolError(f'{self.alias}: [Optno {kind}] invalid format')

        data = dict(
            kind = kind,
            type = self._read_opt_type(kind),
            length = size,
            data = self._read_fileng(size),
        )
        return data

    def _read_mode_unpack(self, size, kind):
        """Read options request unpack process.

        Positional arguemnts:
            * size - int, length of option
            * kind - int, option kind value

        Returns:
            * dict -- extracted option

        Structure of TCP options:
            Octets      Bits        Name                    Discription
              0           0     ip.opt.kind             Kind
              0           0     ip.ts.type.copy         Copied Flag
              0           1     ip.ts.type.class        Option Class
              0           3     ip.ts.type.number       Option Number
              1           8     ip.opt.length           Length
              2          16     ip.opt.data             Kind-specific Data

        """
        if size < 3:
            raise ProtocolError(f'{self.alias}: [Optno {kind}] invalid format')

        data = dict(
            kind = kind,
            type = self._read_opt_type(kind),
            length = size,
            data = self._read_unpack(size),
        )
        return data

    def _read_mode_route(self, size, kind):
        """Read options with route data.

        Positional arguemnts:
            * size - int, length of option
            * kind - int, 7/131/137 (RR/LSR/SSR)

        Returns:
            * dict -- extracted option with route data

        Structure of these options:
            * [RFC 791] Loose Source Route
                +--------+--------+--------+---------//--------+
                |10000011| length | pointer|     route data    |
                +--------+--------+--------+---------//--------+
            * [RFC 791] Strict Source Route
                +--------+--------+--------+---------//--------+
                |10001001| length | pointer|     route data    |
                +--------+--------+--------+---------//--------+
            * [RFC 791] Record Route
                +--------+--------+--------+---------//--------+
                |00000111| length | pointer|     route data    |
                +--------+--------+--------+---------//--------+

            Octets      Bits        Name                    Discription
              0           0     ip.opt.kind             Kind (7/131/137)
              0           0     ip.opt.type.copy        Copied Flag (0)
              0           1     ip.opt.type.class       Option Class (0/1)
              0           3     ip.opt.type.number      Option Number (3/7/9)
              1           8     ip.opt.length           Length
              2          16     ip.opt.pointer          Pointer (≥4)
              3          24     ip.opt.data             Route Data

        """
        if size < 3 or (size - 3) % 4 != 0:
            raise ProtocolError(f'{self.alias}: [Optno {kind}] invalid format')

        _rptr = self._read_unpack(1)

        if _rptr < 4:
            raise ProtocolError(f'{self.alias}: [Optno {kind}] invalid format')

        data = dict(
            kind = kind,
            type = self._read_opt_type(kind),
            length = size,
            pointer = _tptr,
        )

        counter = 4
        address = list()
        endpoint = min(_rptr, size)
        while counter < endpoint:
            counter += 4
            address.append(self._read_ipv4_addr())
        data['ip'] = address or None

        return data

    def _read_mode_qs(self, size, kind):
        """Read Quick Start option.

        Positional arguemnts:
            * size - int, length of option
            * kind - int, 25 (QS)

        Returns:
            * dict -- extracted Quick Start (QS) option

        Structure of Quick-Start (QS) option [RFC 4782]:
            * A Quick-Start Request.
                 0                   1                   2                   3
                 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                |   Option      |  Length=8     | Func. | Rate  |   QS TTL      |
                |               |               | 0000  |Request|               |
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                |                        QS Nonce                           | R |
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            * Report of Approved Rate.
                 0                   1                   2                   3
                 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                |   Option      |  Length=8     | Func. | Rate  |   Not Used    |
                |               |               | 1000  | Report|               |
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                |                        QS Nonce                           | R |
                +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

            Octets      Bits        Name                    Discription
              0           0     ip.qs.kind              Kind (25)
              0           0     ip.qs.type.copy         Copied Flag (0)
              0           1     ip.qs.type.class        Option Class (0)
              0           3     ip.qs.type.number       Option Number (25)
              1           8     ip.qs.length            Length (8)
              2          16     ip.qs.func              Function (0/8)
              2          20     ip.qs.rate              Rate Request / Report (in Kbps)
              3          24     ip.qs.ttl               QS TTL / None
              4          32     ip.qs.nounce            QS Nounce
              7          62     -                       Reserved (\x00\x00)

        """
        if size != 8:
            raise ProtocolError(f'{self.alias}: [Optno {kind}] invalid format')

        _type = self._read_opt_type(kind)
        _fcrr = self._read_binary(1)
        _func = int(_fcrr[:4], base=2)
        _rate = int(_fcrr[4:], base=2)
        _ttlv = self._read_unpack(1)
        _nonr = self._read_binary(4)
        _qsnn = int(_nonr[:30], base=2)

        if _func != 0 and _func != 8:
            raise ProtocolError(f'{self.alias}: [Optno {kind}] invalid format')

        data = dict(
            kind = kind,
            type = _type,
            length = size,
            func = QS_FUNC.get(_func),
            rate = 40000 * (2 ** _rate) / 1000,
            ttl = None if func else _rate,
            nounce = _qsnn,
        )

        return data

    def _read_mode_ts(self, size, kind):
        """Read Time Stamp option.

        Positional asrguemnts:
            * size - int, length of option
            * kind - int, 68 (TS)

        Returns:
            * dict -- extracted Time Stamp (TS) option

        Structure of Timestamp (TS) option [RFC 791]:
            +--------+--------+--------+--------+
            |01000100| length | pointer|oflw|flg|
            +--------+--------+--------+--------+
            |         internet address          |
            +--------+--------+--------+--------+
            |             timestamp             |
            +--------+--------+--------+--------+
            |                 .                 |
                              .
                              .

            Octets      Bits        Name                    Discription
              0           0     ip.ts.kind              Kind (25)
              0           0     ip.ts.type.copy         Copied Flag (0)
              0           1     ip.ts.type.class        Option Class (0)
              0           3     ip.ts.type.number       Option Number (25)
              1           8     ip.ts.length            Length (≤40)
              2          16     ip.ts.pointer           Pointer (≥5)
              3          24     ip.ts.overflow          Overflow Octets
              3          28     ip.ts.flag              Flag
              4          32     ip.ts.ip                Internet Address
              8          64     ip.ts.timestamp         Timestamp

        """
        if size > 40 or size < 4:
            raise ProtocolError(f'{self.alias}: [Optno {kind}] invalid format')

        _tptr = self._read_unpack(1)
        _oflg = self._read_binary(1)
        _oflw = int(_oflg[:4], base=2)
        _flag = int(_flag[4:], base=2)

        if _tptr < 5:
            raise ProtocolError(f'{self.alias}: [Optno {kind}] invalid format')

        data = dict(
            kind = kind,
            type = self._read_opt_type(kind),
            length = size,
            pointer = _tptr,
            overflow = _oflw,
            flag = _flag,
        )

        endpoint = min(_tptr, size)
        if _flag == 0:
            if (size - 4) % 4 != 0:
                raise ProtocolError(f'{self.alias}: [Optno {kind}] invalid format')
            counter = 5
            timestamp = list()
            while counter < endpoint:
                counter += 4
                time = self._read_unpack(4, lilendian=True)
                timestamp.appned(datetime.datetime.fromtimestamp(time))
            data['timestamp'] = timestamp or None
        elif _flag == 1 or _flag == 3:
            if (size - 4) % 8 != 0:
                raise ProtocolError(f'{self.alias}: [Optno {kind}] invalid format')
            counter = 5
            ipaddress = list()
            timestamp = list()
            while counter < endpoint:
                counter += 8
                ipaddress.append(self._read_ipv4_addr())
                time = self._read_unpack(4, lilendian=True)
                timestamp.append(datetime.datetime.fromtimestamp(time))
            data['ip'] = ipaddress or None
            data['timestamp'] = timestamp or None
        else:
            data['data'] = self._read_fileng(size - 4) or None

        return data

    def _read_mode_tr(self, size, kind):
        """Read Traceroute option.

        Positional arguemnts:
            size - int, length of option
            kind - int, 82 (TR)

        Returns:
            * dict -- extracted Traceroute (TR) option

        Structure of Traceroute (TR) option [RFC 1393][RFC 6814]:
             0               8              16              24
            +-+-+-+-+-+-+-+-+---------------+---------------+---------------+
            |F| C |  Number |    Length     |          ID Number            |
            +-+-+-+-+-+-+-+-+---------------+---------------+---------------+
            |      Outbound Hop Count       |       Return Hop Count        |
            +---------------+---------------+---------------+---------------+
            |                     Originator IP Address                     |
            +---------------+---------------+---------------+---------------+

            Octets      Bits        Name                    Discription
              0           0     ip.tr.kind              Kind (82)
              0           0     ip.tr.type.copy         Copied Flag (0)
              0           1     ip.tr.type.class        Option Class (0)
              0           3     ip.tr.type.number       Option Number (18)
              1           8     ip.tr.length            Length (12)
              2          16     ip.tr.id                ID Number
              4          32     ip.tr.ohc               Outbound Hop Count
              6          48     ip.tr.rhc               Return Hop Count
              8          64     ip.tr.ip                Originator IP Address

        """
        if size != 12:
            raise ProtocolError(f'{self.alias}: [Optno {kind}] invalid format')

        _idnm = self._read_unpack(2)
        _ohcn = self._read_unpack(2)
        _rhcn = self._read_unpack(2)
        _ipad = self._read_ipv4_addr()

        data = dict(
            kind = kind,
            type = self._read_opt_type(kind),
            length = size,
            id = _idnm,
            ohc = _ohcn,
            rhc = _rhcn,
            ip = _ipad,
        )

        return data

    def _read_mode_sec(self, size, kind):
        """Read options with security info.

        Positional arguemnts:
            size - int, length of option
            kind - int, 130 (SEC )/ 133 (ESEC)

        Returns:
            * dict -- extracted option with security info (E/SEC)

        Structure of these options:
            * [RFC 1108] Security (SEC)
                +------------+------------+------------+-------------//----------+
                |  10000010  |  XXXXXXXX  |  SSSSSSSS  |  AAAAAAA[1]    AAAAAAA0 |
                |            |            |            |         [0]             |
                +------------+------------+------------+-------------//----------+
                  TYPE = 130     LENGTH   CLASSIFICATION         PROTECTION
                                               LEVEL              AUTHORITY
                                                                    FLAGS
            * [RFC 1108] Extended Security (ESEC):
                +------------+------------+------------+-------//-------+
                |  10000101  |  000LLLLL  |  AAAAAAAA  |  add sec info  |
                +------------+------------+------------+-------//-------+
                 TYPE = 133      LENGTH     ADDITIONAL      ADDITIONAL
                                           SECURITY INFO     SECURITY
                                            FORMAT CODE        INFO

            Octets      Bits        Name                    Discription
              0           0     ip.sec.kind             Kind (130)
              0           0     ip.sec.type.copy        Copied Flag (1)
              0           1     ip.sec.type.class       Option Class (0)
              0           3     ip.sec.type.number      Option Number (2)
              1           8     ip.sec.length           Length (≥3)
              2          16     ip.sec.level            Classification Level
              3          24     ip.sec.flags            Protection Authority Flags

        """
        if size < 3:
            raise ProtocolError(f'{self.alias}: [Optno {kind}] invalid format')

        _clvl = self._read_binary(1)

        data = dict(
            kind = kind,
            type = self._read_opt_type(kind),
            length = size,
            level = _CLASSIFICATION_LEVEL.get(_clvl, _clvl),
        )

        if size > 3:
            _list = list()
            for counter in range(3, size):
                _flag = self._read_binary(1)
                if (counter < size - 1 and not int(_flag[7], base=2)) \
                        or (counter == size - 1 and int(_flag[7], base=2)):
                    raise ProtocolError(f'{self.alias}: [Optno {kind}] invalid format')

                _dict = dict()
                for (index, bit) in enumerate(_flag[:5]):
                    _auth = _PROTECTION_AUTHORITY.get(index)
                    _dict[_auth] = True if int(bit, base=2) else False
                _list.append(Info(_dict))
            data['flags'] = tuple(_list)

        return data

    def _read_mode_rsralt(self, size, kind):
        """Read Router Alert option.

        Positional arguemnts:
            size - int, length of option
            kind - int, 148 (RTRALT)

        Returns:
            * dict -- extracted Router Alert (RTRALT) option

        Structure of Router Alert (RTRALT) option [RFC 2113]:
            +--------+--------+--------+--------+
            |10010100|00000100|  2 octet value  |
            +--------+--------+--------+--------+

            Octets      Bits        Name                    Discription
              0           0     ip.rsralt.kind          Kind (148)
              0           0     ip.rsralt.type.copy     Copied Flag (1)
              0           1     ip.rsralt.type.class    Option Class (0)
              0           3     ip.rsralt.type.number   Option Number (20)
              1           8     ip.rsralt.length        Length (4)
              2          16     ip.rsralt.alert         Alert
              2          16     ip.rsralt.code          Alert Code

        """
        if size != 4:
            raise ProtocolError(f'{self.alias}: [Optno {kind}] invalid format')

        _code = self._read_unpack(2)

        data = dict(
            kind = kind,
            type = self._read_opt_type(kind),
            length = size,
            alert = _ROUTER_ALERT.get(_code, 'Reserved'),
            code = _code,
        )

        return data



# -*- coding: utf-8 -*-
"""decorator functions

`jspcap.utilities.decorators` contains several useful
decorators, including `seekset` and `beholder`.

"""
import functools
import io
import os

###############################################################################
# from jspcap.protocols.raw import Raw
# from jspcap.foundation.analysis import Analysis
###############################################################################


__all__ = ['seekset', 'seekset_ng', 'beholder', 'beholder_ng']


def seekset(func):
    """[ClassMethod] Read file from start then set back to original."""
    @functools.wraps(func)
    def seekcur(self, *args, **kw):
        seek_cur = self._file.tell()
        self._file.seek(os.SEEK_SET)
        return_ = func(self, *args, **kw)
        self._file.seek(seek_cur, os.SEEK_SET)
        return return_
    return seekcur


def seekset_ng(func):
    """Read file from start then set back to original."""
    @functools.wraps(func)
    def seekcur(file, *args, **kw):
        seek_cur = file.tell()
        file.seek(os.SEEK_SET)
        return_ = func(file, *args, **kw)
        file.seek(seek_cur, os.SEEK_SET)
        return return_
    return seekcur


def beholder(func):
    """[ClassMethod] Behold extraction procedure."""
    @functools.wraps(func)
    def behold(self, proto, length, *args, **kwargs):
        seek_cur = self._file.tell()
        try:
            return func(proto, length, *args, **kwargs)
        except Exception as error:
            from jspcap.protocols.raw import Raw

            self._file.seek(seek_cur, os.SEEK_SET)
            next_ = Raw(io.BytesIO(self._read_fileng(length)), length, error=str(error))
            return False, next_.info, next_.protochain, next_.alias
    return behold


def beholder_ng(func):
    """Behold analysis procedure."""
    @functools.wraps(func)
    def behold(file, length, *args, **kwargs):
        seek_cur = file.tell()
        try:
            return func(file, length, *args, **kwargs)
        except Exception as error:
            from jspcap.foundation.analysis import Analysis
            from jspcap.protocols.raw import Raw

            file.seek(seek_cur, os.SEEK_SET)

            raw = Raw(file, length, error=str(error))
            return Analysis(raw.info, raw.protochain, raw.alias)
    return behold



# -*- coding: utf-8 -*-
"""internet protocol version 6

`jspcap.protocols.internet.ipv6` contains `IPv6` only,
which implements extractor for Internet Protocol version 6
(IPv6), whose structure is described as below.

 0                   1                   2                   3
 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|Version| Traffic Class |           Flow Label                  |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|         Payload Length        |  Next Header  |   Hop Limit   |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                                                               |
+                                                               +
|                                                               |
+                         Source Address                        +
|                                                               |
+                                                               +
|                                                               |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                                                               |
+                                                               +
|                                                               |
+                      Destination Address                      +
|                                                               |
+                                                               +
|                                                               |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

"""
# TODO: Implements IPv6 extension headers.
import collections
import ipaddress

from jspcap.corekit.infoclass import Info
from jspcap.corekit.protochain import ProtoChain
from jspcap.protocols.internet.ip import IP


__all__ = ['IPv6']


# IPv6 Extension Header Types
EXT_HDR = (
    'HOPOPT',       # IPv6 Hop-by-Hop Option
    'IPv6-Route',   # Routing Header for IPv6
    'IPv6-Frag',    # Fragment Header for IPv6
    'ESP',          # Encapsulating Security Payload
    'AH',           # Authentication Header
    'IPv6-NoNxt',   # No Next Header for IPv6
    'IPv6-Opts',    # Destination Options for IPv6 (before routing / upper-layer header)
    'MH',           # Mobility Extension Header for IPv6 (currently without upper-layer header)
    'HIP',          # Host Identity Protocol
    'Shim6',        # Site Multihoming by IPv6 Intermediation
)


class IPv6(IP):
    """This class implements Internet Protocol version 6.

    Properties:
        * name -- str, name of corresponding procotol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding procotol
        * layer -- str, `Internet`
        * length -- int, header length of corresponding protocol
        * protocol -- str, name of next layer protocol
        * protochain -- ProtoChain, protocol chain of current instance
        * src -- str, source IP address
        * dst -- str, destination IP address

    Methods:
        * read_ipv6 -- read Internet Protocol version 6 (IPv6)

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _decode_next_layer -- decode next layer protocol type
        * _import_next_layer -- import next layer protocol extractor
        * _read_ip_hextet -- read first four hextets of IPv6
        * _read_ip_addr -- read IP address

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def name(self):
        """Name of corresponding procotol."""
        return 'Internet Protocol version 6'

    @property
    def length(self):
        """Header length of corresponding protocol."""
        return self._info.hdr_len

    @property
    def protocol(self):
        """Name of next layer protocol."""
        return self._info.next

    ##########################################################################
    # Methods.
    ##########################################################################

    def read_ipv6(self, length):
        """Read Internet Protocol version 6 (IPv6).

        Structure of IPv6 header [RFC 2460]:

             0                   1                   2                   3
             0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |Version| Traffic Class |           Flow Label                  |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |         Payload Length        |  Next Header  |   Hop Limit   |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |                                                               |
            +                                                               +
            |                                                               |
            +                         Source Address                        +
            |                                                               |
            +                                                               +
            |                                                               |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |                                                               |
            +                                                               +
            |                                                               |
            +                      Destination Address                      +
            |                                                               |
            +                                                               +
            |                                                               |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

            Octets      Bits        Name                    Discription
              0           0     ip.version              Version (6)
              0           4     ip.class                Traffic Class
              1          12     ip.label                Flow Label
              4          32     ip.payload              Payload Length (header excludes)
              6          48     ip.next                 Next Header
              7          56     ip.limit                Hop Limit
              8          64     ip.src                  Source Address
              24        192     ip.dst                  Destination Address

        """
        if length is None:
            length = len(self)

        _htet = self._read_ip_hextet()
        _plen = self._read_unpack(2)
        _next = self._read_protos(1)
        _hlmt = self._read_unpack(1)
        _srca = self._read_ip_addr()
        _dsta = self._read_ip_addr()

        ipv6 = dict(
            version = _htet[0],
            tclass = _htet[1],
            label = _htet[2],
            payload = _plen,
            next = _next,
            limit = _hlmt,
            src = _srca,
            dst = _dsta,
        )

        hdr_len = 40
        raw_len = ipv6['payload']
        ipv6['packet'] = self._read_packet(header=hdr_len, payload=raw_len)

        return self._decode_next_layer(ipv6, _next, raw_len)

    ##########################################################################
    # Data models.
    ##########################################################################

    def __init__(self, _file, length=None, **kwargs):
        self._file = _file
        self._info = Info(self.read_ipv6(length))

    def __length_hint__(self):
        return 40

    @classmethod
    def __index__(cls):
        return cls.__name__

    ##########################################################################
    # Utilities.
    ##########################################################################

    def _read_ip_hextet(self):
        """Read first four hextets of IPv6."""
        _htet = self._read_fileng(4).hex()
        _vers = _htet[0]                    # version number (6)
        _tcls = int(_htet[0:2], base=16)    # traffic class
        _flow = int(_htet[2:], base=16)     # flow label

        return (_vers, _tcls, _flow)

    def _read_ip_addr(self):
        """Read IP address."""
        # adlt = []       # list of IPv6 hexadecimal address
        # ctr_ = collections.defaultdict(int)
        #                 # counter for consecutive groups of zero value
        # ptr_ = 0        # start pointer of consecutive groups of zero value
        # last = False    # if last hextet/group is zero value
        # ommt = False    # ommitted flag, since IPv6 address can ommit to `::` only once

        # for _ in range(8):
        #     hex_ = self._read_fileng(2).hex().lstrip('0')

        #     if hex_:    # if hextet is not '', directly append
        #         adlt.append(hex_)
        #         last = False
        #     else:       # if hextet is '', append '0'
        #         adlt.append('0')
        #         if last:    # if last hextet is '', ascend counter
        #             ctr_[ptr_] += 1
        #         else:       # if last hextet is not '', record pointer
        #             ptr_ = _
        #             last = True
        #             ctr_[ptr_] = 1

        # ptr_ = max(ctr_, key=ctr_.get) if ctr_ else 0   # fetch start pointer with longest zero values
        # end_ = ptr_ + ctr_[ptr_]                        # calculate end pointer

        # if ctr_[ptr_] > 1:      # only ommit if zero values are in a consecutive group
        #     del adlt[ptr_:end_] # remove zero values

        #     if ptr_ == 0 and end_ == 8:     # insert `::` if IPv6 unspecified address (::)
        #         adlt.insert(ptr_, '::')
        #     elif ptr_ == 0 or end_ == 8:    # insert `:` if zero values are from start or at end
        #         adlt.insert(ptr_, ':')
        #     else:                           # insert '' otherwise
        #         adlt.insert(ptr_, '')

        # addr = ':'.join(adlt)
        # return addr
        return ipaddress.ip_address(self._read_fileng(16))

    def _decode_next_layer(self, ipv6, proto=None, length=None):
        """Decode next layer extractor.

        Positional arguments:
            * ipv6 -- dict, info buffer
            * proto -- str, next layer protocol name
            * length -- int, valid (not padding) length

        Returns:
            * dict -- current protocol with next layer extracted

        """
        # recurse if next header is an extensive header
        hdr_len = 40                # header length
        raw_len = ipv6['payload']   # payload length
        while proto in EXT_HDR:
            # keep original data after fragment header
            if proto == 'IPv6-Frag':
                ipv6['fragment'] = self._read_packet(header=hdr_len, payload=raw_len)

            # directly break when No Next Header ocuurs
            if proto == 'IPv6-NoNxt':
                proto = None
                break

            # make protocol name
            flag, info, chain, alias = self._import_next_layer(proto, version=6, extension=True)
            name = proto.replace('IPv6-', '').lower() if flag else 'raw'
            ipv6[name] = info

            # record protocol name
            self._protos = ProtoChain(name, chain, alias)
            if not flag:
                proto = None
                break
            proto = info.next

            # update header & payload length
            hdr_len += info.length
            raw_len -= info.length

        # record real header & payload length (headers exclude)
        ipv6['hdr_len'] = hdr_len
        ipv6['raw_len'] = raw_len

        # update next header
        ipv6['proto'] = None if proto in EXT_HDR else proto
        return super()._decode_next_layer(ipv6, proto, raw_len)



# -*- coding: utf-8 -*-
"""(inverse) address resolution protocol

`jspcap.protocols.link.arp` contains `ARP` only,
which implements extractor for (Inverse) Address Resolution
Protocol (ARP/InARP), whose structure is described as below.

Octets      Bits        Name                    Discription
  0           0     arp.htype               Hardware Type
  2          16     arp.ptype               Protocol Type
  4          32     arp.hlen                Hardware Address Length
  5          40     arp.plen                Protocol Address Length
  6          48     arp.oper                Operation
  8          64     arp.sha                 Sender Hardware Address
  14        112     arp.spa                 Sender Protocol Address
  18        144     arp.tha                 Target Hardware Address
  24        192     arp.tpa                 Target Protocol Address

"""
import collections
import ipaddress
import textwrap

from jspcap.corekit.infoclass import Info
from jspcap.protocols.internet.internet import ETHERTYPE
from jspcap.protocols.link.link import Link, LINKTYPE


__all__ = ['ARP']


# Operation Codes [RFC 826][RFC 5494]
OPER = {
    0 : 'Reserved',         # [RFC 5494]
    1 : 'REQUEST',          # [RFC 826][RFC 5227]
    2 : 'REPLY',            # [RFC 826][RFC 5227]
    3 : 'Request Reverse',  # [RFC 826][RFC 5227]
    4 : 'Reply Reverse',    # [RFC 903]
    5 : 'DRARP-Request',    # [RFC 1931]
    6 : 'DRARP-Reply',      # [RFC 1931]
    7 : 'DRARP-Error',      # [RFC 1931]
    8 : 'InARP-Request',    # [RFC 2390]
    9 : 'InARP-Reply',      # [RFC 2390]
   10 : 'ARP-NAK',          # [RFC 1577]
65535 : 'Reserved',         # [RFC 5494]
}


class ARP(Link):
    """This class implements all protocols in ARP family.

    - Address Resolution Protocol (ARP) [RFC 826]
    - Reverse Address Resolution Protocol (RARP) [RFC 903]
    - Dynamic Reverse Address Resolution Protocol (DRARP) [RFC 1931]
    - Inverse Address Resolution Protocol (InARP) [RFC 2390]

    Properties:
        * name -- str, name of corresponding protocol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding protocol
        * layer -- str, `Link`
        * length -- int, header length of corresponding protocol
        * protocol -- str, name of next layer protocol
        * protochain -- ProtoChain, protocol chain of current instance
        * src -- tuple<str, str>, sender hardware & protocol address
        * dst -- tuple<str, str>, target hardware & protocol address
        * type -- tuple<str, str>, hardware & protocol type

    Methods:
        * read_arp -- read Address Resolution Protocol

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance
        * _acnm -- str, acronym of corresponding protocol
        * _name -- str, name of corresponding protocol

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _decode_next_layer -- decode next layer protocol type
        * _import_next_layer -- import next layer protocol extractor
        * _read_addr_resolve -- resolve MAC address according to protocol
        * _read_proto_resolve -- solve IP address according to protocol

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def name(self):
        """Name of current protocol."""
        return self._name

    @property
    def alias(self):
        """Acronym of corresponding protocol."""
        return self._acnm

    @property
    def length(self):
        """Header length of current protocol."""
        return self._info.len

    @property
    def src(self):
        """Sender hardware & protocol address."""
        return (self._info.sha, self._info.spa)

    @property
    def dst(self):
        """Target hardware & protocol address."""
        return (self._info.tha, self._info.tpa)

    @property
    def type(self):
        """Hardware & protocol type."""
        return (self._info.htype, self._info.ptype)

    ##########################################################################
    # Methods.
    ##########################################################################

    def read_arp(self, length):
        """Read Address Resolution Protocol.

        Structure of ARP header [RFC 826]:
            Octets      Bits        Name                    Discription
              0           0     arp.htype               Hardware Type
              2          16     arp.ptype               Protocol Type
              4          32     arp.hlen                Hardware Address Length
              5          40     arp.plen                Protocol Address Length
              6          48     arp.oper                Operation
              8          64     arp.sha                 Sender Hardware Address
              14        112     arp.spa                 Sender Protocol Address
              18        144     arp.tha                 Target Hardware Address
              24        192     arp.tpa                 Target Protocol Address

        """
        if length is None:
            length = len(self)

        _hwty = self._read_unpack(2)
        _ptty = self._read_unpack(2)
        _hlen = self._read_unpack(1)
        _plen = self._read_unpack(1)
        _oper = self._read_unpack(2)
        _shwa = self._read_addr_resolve(_hlen, _hwty)
        _spta = self._read_proto_resolve(_plen, _ptty)
        _thwa = self._read_addr_resolve(_hlen, _hwty)
        _tpta = self._read_proto_resolve(_plen, _ptty)

        if _oper in (5, 6, 7):
            self._acnm = 'DRARP'
            self._name = 'Dynamic Reverse Address Resolution Protocol'
        elif _oper in (8, 9):
            self._acnm = 'InARP'
            self._name = 'Inverse Address Resolution Protocol'
        elif _oper in (3, 4):
            self._acnm = 'RARP'
            self._name = 'Reverse Address Resolution Protocol'
        else:
            self._acnm = 'ARP'
            self._name = 'Address Resolution Protocol'

        arp = dict(
            htype = LINKTYPE.get(_hwty),
            ptype = ETHERTYPE.get(_ptty),
            hlen = _hlen,
            plen = _plen,
            oper = OPER.get(_oper),
            sha = _shwa,
            spa = _spta,
            tha = _thwa,
            tpa = _tpta,
            len = 8 + _hlen * 2 + _plen * 2,
        )

        length -= arp['len']
        arp['packet'] = self._read_packet(header=arp['len'], payload=length)

        return self._decode_next_layer(arp, None, length)

    ##########################################################################
    # Data models.
    ##########################################################################

    def __init__(self, _file, length=None, **kwargs):
        self._file = _file
        self._info = Info(self.read_arp(length))

    def __length_hint__(self):
        return 28

    @classmethod
    def __index__(cls):
        return ('ARP', 'InARP')

    ##########################################################################
    # Utilities.
    ##########################################################################

    def _read_addr_resolve(self, length, htype):
        """Resolve MAC address according to protocol.

        Positional arguments:
            * length -- int, hardware address length
            * htype -- int, hardware type

        Returns:
            * str -- MAC address

        """
        if htype == 1:  # Ethernet
            _byte = self._read_fileng(6)
            _addr = '-'.join(textwrap.wrap(_byte.hex(), 2))
        else:
            _addr = self._read_fileng(length)
        return _addr

    def _read_proto_resolve(self, length, ptype):
        """Resolve IP address according to protocol.

        Positional arguments:
            * length -- int, protocol address length
            * hptype -- int, protocol type

        Returns:
            * str -- IP address

        """
        # if ptype == '0800':     # IPv4
        #     _byte = self._read_fileng(4)
        #     _addr = '.'.join([str(_) for _ in _byte])
        # elif ptype == '86dd':   # IPv6
        #     adlt = []       # list of IPv6 hexadecimal address
        #     ctr_ = collections.defaultdict(int)
        #                     # counter for consecutive groups of zero value
        #     ptr_ = 0        # start pointer of consecutive groups of zero value
        #     last = False    # if last hextet/group is zero value
        #     ommt = False    # ommitted flag, since IPv6 address can ommit to `::` only once

        #     for index in range(8):
        #         hex_ = self._read_fileng(2).hex().lstrip('0')

        #         if hex_:    # if hextet is not '', directly append
        #             adlt.append(hex_)
        #             last = False
        #         else:       # if hextet is '', append '0'
        #             adlt.append('0')
        #             if last:    # if last hextet is '', ascend counter
        #                 ctr_[ptr_] += 1
        #             else:       # if last hextet is not '', record pointer
        #                 ptr_ = index
        #                 last = True
        #                 ctr_[ptr_] = 1

        #     ptr_ = max(ctr_, key=ctr_.get) if ctr_ else 0   # fetch start pointer with longest zero values
        #     end_ = ptr_ + ctr_[ptr_]                        # calculate end pointer

        #     if ctr_[ptr_] > 1:      # only ommit if zero values are in a consecutive group
        #         del adlt[ptr_:end_] # remove zero values

        #         if ptr_ == 0 and end_ == 8:     # insert `::` if IPv6 unspecified address (::)
        #             adlt.insert(ptr_, '::')
        #         elif ptr_ == 0 or end_ == 8:    # insert `:` if zero values are from start or at end
        #             adlt.insert(ptr_, ':')
        #         else:                           # insert '' otherwise
        #             adlt.insert(ptr_, '')

        #     _addr = ':'.join(adlt)
        # else:
        #     _addr = self._read_fileng(length)
        # return _addr
        if ptype == '0800':     # IPv4
            return ipaddress.ip_address(self._read_fileng(4))
        elif ptype == '86dd':   # IPv6
            return ipaddress.ip_address(self._read_fileng(16))
        else:
            return self._read_fileng(length)



# -*- coding: utf-8 -*-
"""reassembly packets and datagrams

`jspcap.reassembly` bases on algorithms described in
[`RFC 815`](https://tools.ietf.org/html/rfc815),
implements datagram reassembly of IP and TCP packets.

"""
# Base Class for Reassembly
from jspcap.reassembly.reassembly import Reassembly
from jspcap.reassembly.ip import IP_Reassembly

# Reassembly for IP
from jspcap.reassembly.ipv4 import IPv4_Reassembly
from jspcap.reassembly.ipv6 import IPv6_Reassembly

# Reassembly for TCP
from jspcap.reassembly.tcp import TCP_Reassembly


__all__ = [
    'IPv4_Reassembly', 'IPv6_Reassembly',   # IP Reassembly
    'TCP_Reassembly',                       # TCP Reassembly
]



"""hypertext transfer protocol version 2

``jspcap.protocols.application.httpv2`` contains ``HTTPv2``
only, which implements extractor for Hypertext Transfer
Protocol version 2 (HTTP/2), whose structure is described
as below.

+-----------------------------------------------+
|                 Length (24)                   |
+---------------+---------------+---------------+
|   Type (8)    |   Flags (8)   |
+-+-------------+---------------+-------------------------------+
|R|                 Stream Identifier (31)                      |
+=+=============================================================+
|                   Frame Payload (0...)                      ...
+---------------------------------------------------------------+

"""
# Hypertext Transfer Protocol version 2
# Analyser for HTTP/2 packets


from jspcap.exceptions import ProtocolError
from jspcap.utilities import Info
from jspcap.protocols.application.http import HTTP


class HTTPv2(HTTP):
    """This class implements Hypertext Transfer Protocol version 2.

    Properties:
        * name -- str, name of corresponding procotol
        * info -- Info, info dict of current instance
        * layer -- str, `Application`
        * protocol -- str, name of next layer protocol
        * protochain -- ProtoChain, protocol chain of current instance

    Methods:
        * read_http -- read Hypertext Transfer Protocol version 2 (HTTP/2)

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries

    """
    ##########################################################################
    # Methods.
    ##########################################################################

    def read_http(self, length):
        """Read Hypertext Transfer Protocol version 2.

        Structure of HTTP/2 packet [RFC 7230]:
            +-----------------------------------------------+
            |                 Length (24)                   |
            +---------------+---------------+---------------+
            |   Type (8)    |   Flags (8)   |
            +-+-------------+---------------+-------------------------------+
            |R|                 Stream Identifier (31)                      |
            +=+=============================================================+
            |                   Frame Payload (0...)                      ...
            +---------------------------------------------------------------+

        """
        _plen = self._read_binary(3)
        _type = self._read_unpack(1)
        _flag = self._read_binary(1)
        _stid = self._read_binary(4)



# -*- coding: utf-8 -*-
"""validation utilities

`jspcap.utilities.validations` contains functions to validate
arguments for functions and classes. It was first used in
[`jsntlib`](https://github.com/JarryShaw/jspcapy) as
validators.

"""
import inspect
import io
import numbers

from jspcap.utilities.exceptions import BoolError, BytesError, BytearrayError, \
        DictError, DigitError, FragmentError, IntError, IOObjError, ListError, \
        RealError, TupleError, InfoError, PacketError


__all__ = [
    'int_check', 'real_check', 'complex_check', 'number_check',
    'bool_check', 'bytes_check', 'bytearray_check', 'str_check',
    'list_check', 'dict_check', 'tuple_check', 'io_check',
    'frag_check', 'pkt_check', 'info_check',
]


def int_check(*args, func=None):
    """Check if arguments are integrals."""
    func = func or inspect.stack()[2][3]
    for var in args:
        if not isinstance(var, numbers.Integral):
            name = type(var).__name__
            raise ComplexError(f'Function {func} expected integral number, {name} got instead.')


def real_check(*args, func=None):
    """Check if arguments are real numbers."""
    func = func or inspect.stack()[2][3]
    for var in args:
        if not isinstance(var, numbers.Real):
            name = type(var).__name__
            raise ComplexError(f'Function {func} expected real number, {name} got instead.')


def complex_check(*args, func=None):
    """Check if arguments are complex numbers."""
    func = func or inspect.stack()[2][3]
    for var in args:
        if not isinstance(var, numbers.Complex):
            name = type(var).__name__
            raise ComplexError(f'Function {func} expected complex number, {name} got instead.')


def number_check(*args, func=None):
    """Check if arguments are numbers."""
    func = func or inspect.stack()[2][3]
    for var in args:
        if not isinstance(var, numbers.Number):
            name = type(var).__name__
            raise DigitError(f'Function {func} expected number, {name} got instead.')


def bytes_check(*args, func=None):
    """Check if arguments are bytes type."""
    func = func or inspect.stack()[2][3]
    for var in args:
        if not isinstance(var, bytes):
            name = type(var).__name__
            raise BytesError(f'Function {func} expected bytes, {name} got instead.')


def bytearray_check(*args, func=None):
    """Check if arguments are bytearray type."""
    func = func or inspect.stack()[2][3]
    for var in args:
        if not isinstance(var, bytearray):
            name = type(var).__name__
            raise BytearrayError(f'Function {func} expected bytearray, {name} got instead.')


def str_check(*args, func=None):
    """Check if arguments are str type."""
    func = func or inspect.stack()[2][3]
    for var in args:
        if not isinstance(var, str):
            name = type(var).__name__
            raise StringError(f'Function {func} expected str, {name} got instead.')


def bool_check(*args, func=None):
    """Check if arguments are bytes type."""
    func = func or inspect.stack()[2][3]
    for var in args:
        if not isinstance(var, bool):
            name = type(var).__name__
            raise BoolError(f'Function {func} expected bool, {name} got instead.')


def list_check(*args, func=None):
    """Check if arguments are list type."""
    func = func or inspect.stack()[2][3]
    for var in args:
        if not isinstance(var, list):
            name = type(var).__name__
            raise ListError(f'Function {func} expected list, {name} got instead.')


def dict_check(*args, func=None):
    """Check if arguments are dict type."""
    func = func or inspect.stack()[2][3]
    for var in args:
        if not isinstance(var, dict):
            name = type(var).__name__
            raise DictError(f'Function {func} expected dict, {name} got instead.')


def tuple_check(*args, func=None):
    """Check if arguments are tuple type."""
    func = func or inspect.stack()[2][3]
    for var in args:
        if not isinstance(var, tuple):
            name = type(var).__name__
            raise TupleError(f'Function {func} expected tuple, {name} got instead.')


def io_check(*args, func=None):
    """Check if arguments are file-like object."""
    func = func or inspect.stack()[2][3]
    for var in args:
        if not isinstance(var, io.IOBase):
            name = type(var).__name__
            raise IOObjError(f'Function {func} expected file-like object, {name} got instead.')

def info_check(*args, func=None):
    """Check if arguments are Info instance."""
    from jspcap.corekit.infoclass import Info

    func = func or inspect.stack()[2][3]
    for var in args:
        if not isinstance(var, Info):
            name = type(var).__name__
            raise IOObjError(f'Function {func} expected Info instance, {name} got instead.')


def frag_check(*args, protocol, func=None):
    """Check if arguments are valid fragments."""
    func = func or inspect.stack()[2][3]
    if 'IP' in protocol:
        _ip_frag_check(*args, func=func)
    elif 'TCP' in protocol:
        _tcp_frag_check(*args, func=func)
    else:
        raise FragmentError(f'Unknown fragmented protocol {protocol}.')


def _ip_frag_check(*args, func=None):
    """Check if arguments are valid IP fragments."""
    func = func or inspect.stack()[2][3]
    for var in args:
        dict_check(var, func=func)
        bufid = var.get('bufid')
        bool_check(var.get('mf'), func=func)
        str_check(bufid[0], bufid[1], bufid[3], func=func)
        bytearray_check(var.get('header'), var.get('payload'), func=func)
        int_check(bufid[2], var.get('num'), var.get('fo'), var.get('ihl'), var.get('tl'), func=func)


def _tcp_frag_check(*args, func=None):
    """Check if arguments are valid TCP fragments."""
    func = func or inspect.stack()[2][3]
    for var in args:
        dict_check(var, func=func)
        bufid = var.get('bufid')
        str_check(bufid[0], bufid[1], func=func)
        bytearray_check(var.get('payload'), func=func)
        bool_check(var.get('syn'), var.get('fin'), func=func)
        int_check(bufid[2], bufid[3], var.get('num'), var.get('ack'), var.get('dsn'),
                    var.get('first'), var.get('last'), var.get('len'), func=func)


def pkt_check(*args, func=None):
    """Check if arguments are valid packets."""
    func = func or inspect.stack()[2][3]
    for var in args:
        dict_check(var, func=func)
        info_check(var.get('frame'), func=func)
        real_check(vat.get('timestamp'), func=func)
        bool_check(var.get('syn'), var.get('fin'), func=func)
        str_check(var.get('protocol'), var.get('src'), var.get('dst'), func=func)
        int_check(var.get('srcport'), var.get('dstport'), var.get('index'), func=func)


###############################################################################
# Test Codes
# 
# func = sys._getframe().f_back.f_code.co_name
# spec = inspect.getfullargspec(test)
# argv = spec.args + spec.kwonlyargs
# for index, var in enumerate(args):
#     if not isinstance(var, numbers.Integral):
#         raise IntError( f'Function {func.__name__} expected argument {argv[index]} '
#                         f'to be integral number, {type(var).__name__} got instead.' )
# 
###############################################################################



# -*- coding: utf-8 -*-
"""authentication header

`jspcap.protocols.internet.ah` contains `AH` only,
which implements extractor for Authentication Header (AH),
whose structure is described as below.

 0                   1                   2                   3
 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
| Next Header   |  Payload Len  |          RESERVED             |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                 Security Parameters Index (SPI)               |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                    Sequence Number Field                      |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                                                               |
+                Integrity Check Value-ICV (variable)           |
|                                                               |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

"""
from jspcap.corekit.infoclass import Info
from jspcap.protocols.internet.ipsec import IPsec
from jspcap.utilities.exceptions import VersionError, ProtocolError


__all__ = ['AH']


class AH(IPsec):
    """This class implements Authentication Header.

    Properties:
        * name -- str, name of corresponding procotol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding procotol
        * layer -- str, `Internet`
        * length -- int, header length of corresponding protocol
        * protocol -- str, name of next layer protocol
        * protochain -- ProtoChain, protocol chain of current instance

    Methods:
        * read_ah -- read Authentication Header (AH)

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _decode_next_layer -- decode next layer protocol type
        * _import_next_layer -- import next layer protocol extractor

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def name(self):
        """Name of corresponding procotol."""
        return 'Authentication Header'

    @property
    def length(self):
        """Info dict of current instance."""
        return self._info.length

    @property
    def protocol(self):
        """Name of next layer protocol."""
        return self._info.next

    ##########################################################################
    # Methods.
    ##########################################################################

    def read_ah(self, length, version, extension):
        """Read Authentication Header.

        Structure of AH header [RFC 4302]:
             0                   1                   2                   3
             0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            | Next Header   |  Payload Len  |          RESERVED             |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |                 Security Parameters Index (SPI)               |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |                    Sequence Number Field                      |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |                                                               |
            +                Integrity Check Value-ICV (variable)           |
            |                                                               |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

            Octets      Bits        Name                    Discription
              0           0     ah.next                 Next Header
              1           8     ah.length               Payload Length
              2          16     -                       Reserved (must be zero)
              4          32     ah.spi                  Security Parameters Index (SPI)
              8          64     ah.seq                  Sequence Number Field
              12         96     ah.icv                  Integrity Check Value (ICV)

        """
        if length is None:
            length = len(self)

        _next = self._read_protos(1)
        _plen = self._read_unpack(1)
        _resv = self._read_fileng(2)
        _scpi = self._read_unpack(4)
        _dsnf = self._read_unpack(4)

        # ICV length & value
        _tlen = 20 + _plen * 4
        _vlen = _tlen - 12
        _chkv = self._read_fileng(_vlen)

        ah = dict(
            next = _next,
            length = _tlen,
            spi = _scpi,
            seq = dsnf,
            icv = _chkv,
        )

        if version == 6:
            _plen = 8 - (_vlen % 8)
        elif version == 4:
            _plen = 4 - (_tlen % 4)
        else:
            raise VersionError(f'Unknown IP version {version}')

        if _plen:   # explicit padding in need
            padding = self._read_binary(_plen)
            if any((int(bit, base=2) for bit in padding)):
                raise ProtocolError(f'{self.alias}: invalid format')

        length -= ah['length']
        ah['packet'] = self._read_packet(header=ah['length'], payload=length)

        if extension:
            self._protos = None
            return ah
        return self._decode_next_layer(ah, _next, length)

    ##########################################################################
    # Data models.
    ##########################################################################

    def __init__(self, _file, length=None, *, version=4, extension=False, **kwargs):
        self._file = _file
        self._info = Info(self.read_ah(length, version, extension))

    def __length_hint__(self):
        return 20

    @classmethod
    def __index__(cls):
        return cls.__name__



# -*- coding: utf-8 -*-
"""hypertext transfer protocol

`jspcap.protocols.application.http` contains `HTTP`
only, which is a base class for Hypertext Transfer
Protocol (HTTP) protocol family, eg. HTTP/1.*, HTTP/2.

"""
import chardet

from jspcap.corekit.infoclass import Info
from jspcap.protocols.application.application import Application
from jspcap.utilities.exceptions import UnsupportedCall, ProtocolError


__all__ = ['HTTP']


class HTTP(Application):
    """This class implements all protocols in HTTP family.

    - Hypertext Transfer Protocol (HTTP/1.1) [RFC 7230]
    - Hypertext Transfer Protocol version 2 (HTTP/2) [RFC 7540]

    Properties:
        * name -- str, name of corresponding procotol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding procotol
        * layer -- str, `Application`
        * protocol -- str, name of next layer protocol
        * protochain -- ProtoChain, protocol chain of current instance

    Methods:
        * read_http -- read Hypertext Transfer Protocol (HTTP)

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _make_protochain -- make ProtoChain instance for corresponding protocol

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def name(self):
        """Name of current protocol."""
        return 'Hypertext Transfer Protocol'

    @property
    def length(self):
        """Deprecated."""
        raise UnsupportedCall(f"'{self.__class__.__name__}' object has no attribute 'length'")

    ##########################################################################
    # Data models.
    ##########################################################################

    def __init__(self, _file, length=None, **kwargs):
        self._file = _file
        self._info = Info(self.read_http(length))
        self._make_protochain()

    @classmethod
    def __index__(cls):
        return ('HTTPv1', 'HTTPv2')



# -*- coding: utf-8 -*-
"""reassembly IPv4 fragments

`jspcap.reassembly.ipv4` contains `IPv4_Reassembly`
only, which reconstructs fragmented IPv4 packets back to
origin. The following algorithm implementment is based on IP
reassembly procedure introduced in RFC 791, using
`RCVBT` (fragment receivedbit table). Though another
algorithm is explained in RFC 815, replacing `RCVBT`,
however, this implementment still used the elder one.
And here is the pseudo-code:

Notations:

    FO    - Fragment Offset
    IHL   - Internet Header Length
    MF    - More Fragments flag
    TTL   - Time To Live
    NFB   - Number of Fragment Blocks
    TL    - Total Length
    TDL   - Total Data Length
    BUFID - Buffer Identifier
    RCVBT - Fragment Received Bit Table
    TLB   - Timer Lower Bound

Algorithm:

    DO {
        BUFID <- source|destination|protocol|identification;

        IF (FO = 0 AND MF = 0) {
            IF (buffer with BUFID is allocated) {
                flush all reassembly for this BUFID;
                Submit datagram to next step;
                DONE.
            }
        }

        IF (no buffer with BUFID is allocated) {
            allocate reassembly resources with BUFID;
            TIMER <- TLB;
            TDL <- 0;
            put data from fragment into data buffer with BUFID
                [from octet FO*8 to octet (TL-(IHL*4))+FO*8];
            set RCVBT bits [from FO to FO+((TL-(IHL*4)+7)/8)];
        }

        IF (MF = 0) {
            TDL <- TL-(IHL*4)+(FO*8)
        }

        IF (FO = 0) {
            put header in header buffer
        }

        IF (TDL # 0 AND all RCVBT bits [from 0 to (TDL+7)/8] are set) {
            TL <- TDL+(IHL*4)
            Submit datagram to next step;
            free all reassembly resources for this BUFID;
            DONE.
        }

        TIMER <- MAX(TIMER,TTL);

    } give up until (next fragment or timer expires);

    timer expires: {
        flush all reassembly with this BUFID;
        DONE.
    }

"""
from jspcap.reassembly.ip import IP_Reassembly


__all__ = ['IPv4_Reassembly']


class IPv4_Reassembly(IP_Reassembly):
    """Reassembly for IPv4 payload.

    Usage:
        >>> from jspcap.reassembly import IPv4_Reassembly
        # Initialise instance:
        >>> ipv4_reassembly = IPv4_Reassembly()
        # Call reassembly:
        >>> ipv4_reassembly(packet_dict)
        # Fetch result:
        >>> result = ipv4_reassembly.datagram

    Properties:
        * name -- str, protocol of current packet
        * count -- int, total number of reassembled packets
        * datagram -- tuple, reassembled datagram, which structure may vary
                        according to its protocol
        * protocol -- str, protocol of current reassembly object

    Methods:
        * reassembly -- perform the reassembly procedure
        * submit -- submit reassembled payload
        * fetch -- fetch datagram
        * index -- return datagram index
        * run -- run automatically

    Attributes:
        * _strflg -- bool, stirct mode flag
        * _buffer -- dict, buffer field
        * _dtgram -- tuple, reassembled datagram

    Terminology:
        * packet_dict = dict(
            bufid = tuple(
                ipv4.src,                   # source IP address
                ipv4.dst,                   # destination IP address
                ipv4.id,                    # identification
                ipv4.proto,                 # payload protocol type
            ),
            num = frame.number,             # original packet range number
            fo = ipv4.frag_offset,          # fragment offset
            ihl = ipv4.hdr_len,             # internet header length
            mf = ipv4.flags.mf,             # more fragment flag
            tl = ipv4.len,                  # total length, header includes
            header = ipv4.header,           # raw bytearray type header
            payload = ipv4.payload,         # raw bytearray type payload
          )
        * (tuple) datagram
            |--> (dict) data
            |       |--> 'NotImplemented' : (bool) True --> implemented
            |       |--> 'index' : (tuple) packet numbers
            |       |                |--> (int) original packet range number
            |       |--> 'packet' : (bytes/None) reassembled IPv4 packet
            |--> (dict) data
            |       |--> 'NotImplemented' : (bool) False --> not implemented
            |       |--> 'index' : (tuple) packet numbers
            |       |                |--> (int) original packet range number
            |       |--> 'header' : (bytes/None) IPv4 header
            |       |--> 'payload' : (tuple/None) partially reassembled IPv4 payload
            |                        |--> (bytes/None) IPv4 payload fragment
            |--> (dict) data ...
        * (dict) buffer --> memory buffer for reassembly
            |--> (tuple) BUFID : (dict)
            |       |--> ipv4.src    |
            |       |--> ipc4.dst    |
            |       |--> ipv4.id     |
            |       |--> ipv4.proto  |
            |                        |--> 'TDL' : (int) total data length
            |                        |--> RCVBT : (bytearray) fragment received bit table
            |                        |               |--> (bytes) b\x00' not received
            |                        |               |--> (bytes) b\x01' received
            |                        |               |--> (bytes) ...
            |                        |--> 'index' : (list) list of reassembled packets
            |                        |               |--> (int) packet range number
            |                        |--> 'header' : (bytearray) header buffer
            |                        |--> 'datagram' : (bytearray) data buffer, holes set to b'\x00'
            |--> (tuple) BUFID ...

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def name(self):
        """Protocol of current packet."""
        return 'Internet Protocol version 4'

    @property
    def protocol(self):
        """Protocol of current reassembly object."""
        return 'IPv4'
    



# -*- coding: utf-8 -*-
"""802.1Q customer VLAN tag type

`jspcap.protocols.link.vlan` contains `VLAN`
only, which implements extractor for 802.1QCustomer
VLAN Tag Type, whose structure is described as below.

Octets      Bits        Name                    Discription
  1           0     vlan.tci                Tag Control Information
  1           0     vlan.tci.pcp            Priority Code Point
  1           3     vlan.tci.dei            Drop Eligible Indicator
  1           4     vlan.tci.vid            VLAN Identifier
  3          24     vlan.type               Protocol (Internet Layer)

"""
from jspcap.corekit.infoclass import Info
from jspcap.protocols.link.link import Link

__all__ = ['VLAN']


# priority levels defined in IEEE 802.1p
_PCP = {
    '000' : 'BE',   # 1 - Best effort (default)
    '001' : 'BK',   # 0 - Background (lowest)
    '010' : 'EE',   # 2 - Excellent effort
    '011' : 'CA',   # 3 - Critical applications
    '100' : 'VI',   # 4 - Video, < 100 ms latency and jitter
    '101' : 'VO',   # 5 - Voice, < 10 ms latency and jitter
    '110' : 'IC',   # 6 - Internetwork control
    '111' : 'NC',   # 7 - Network control (highest)
}


class VLAN(Link):
    """This class implements 802.1Q Customer VLAN Tag Type.

    Properties:
        * name -- str, name of corresponding protocol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding protocol
        * layer -- str, `Link`
        * length -- int, header length of corresponding protocol
        * protocol -- str, next layer protocol
        * protochain -- ProtoChain, protocol chain of current instance

    Methods:
        * read_vlan -- read 802.1Q Customer VLAN Tag Type

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _decode_next_layer -- decode next layer protocol type
        * _import_next_layer -- import next layer protocol extractor

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def name(self):
        """Name of current protocol."""
        return '802.1Q Customer VLAN Tag Type'

    @property
    def alias(self):
        """Acronym of corresponding protocol."""
        return '802.1Q'

    @property
    def length(self):
        """Header length of current protocol."""
        return 4

    @property
    def protocol(self):
        """Name of next layer protocol."""
        return self._info.type

    ##########################################################################
    # Methods.
    ##########################################################################

    def read_vlan(self, length):
        """Read 802.1Q Customer VLAN Tag Type.

        Structure of 802.1Q Customer VLAN Tag Type [RFC 7042]:
            Octets      Bits        Name                    Discription
              1           0     vlan.tci                Tag Control Information
              1           0     vlan.tci.pcp            Priority Code Point
              1           3     vlan.tci.dei            Drop Eligible Indicator
              1           4     vlan.tci.vid            VLAN Identifier
              3          24     vlan.type               Protocol (Internet Layer)

        """
        if length is None:
            length = len(self)

        _tcif = self._read_binary(2)
        _type = self._read_protos(2)

        vlan = dict(
            tci = dict(
                pcp = _PCP.get(_tcif[:3]),
                dei = True if _tcif[3] else False,
                vid = int(_tcif[4:], base=2),
            ),
            type = _type,
        )

        length -= 4
        vlan['packet'] = self._read_packet(header=4, payload=length)

        return self._decode_next_layer(vlan, _type, length)

    ##########################################################################
    # Data models.
    ##########################################################################

    def __init__(self, _file, length=None, **kwargs):
        self._file = _file
        self._info = Info(self.read_vlan(length))

    def __length_hint__(self):
        return 4



# -*- coding: utf-8 -*-
"""root link layer protocol

`jspcap.protocols.link.link` contains both `LINKTYPE`
and `Link`. The former is a dictionary of link layer header
type values, registered in IANA. And the latter is a base
class for link layer protocols, eg. ARP/InARP, Ethernet,
L2TP, OSPF, RARP/DRARP and etc.

"""
import io

from jspcap.protocols.internet.internet import ETHERTYPE
from jspcap.protocols.protocol import Protocol


__all__ = ['Link', 'LINKTYPE']


# Link-Layer Header Type Values
LINKTYPE = {
    0 : 'Null',     # BSD loopback encapsulation
    1 : 'Ethernet', # IEEE 802.3 Ethernet
  101 : 'Raw',      # Raw IP
  228 : 'IPv4',     # Raw IPv4
  229 : 'IPv6',     # Raw IPv6
  248 : 'SCTP',     # SCTP packets
}


class Link(Protocol):
    """Abstract base class for link layer protocol family.

    Properties:
        * name -- str, name of corresponding protocol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding protocol
        * layer -- str, `Link`
        * length -- int, header length of corresponding protocol
        * protocol -- str, name of next layer protocol
        * protochain -- ProtoChain, protocol chain of current instance

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _decode_next_layer -- decode next layer protocol type
        * _import_next_layer -- import next layer protocol extractor

    """
    __layer__ = 'Link'

    ##########################################################################
    # Properties.
    ##########################################################################

    # protocol layer
    @property
    def layer(self):
        """Protocol layer."""
        return self.__layer__

    ##########################################################################
    # Utilities.
    ##########################################################################

    def _read_protos(self, size):
        """Read next layer protocol type.

        Positional arguments:
            * size  -- int, buffer size

        Returns:
            * str -- next layer's protocol name

        """
        _byte = self._read_unpack(size)
        _prot = ETHERTYPE.get(_byte)
        return _prot

    def _import_next_layer(self, proto, length):
        """Import next layer extractor.

        Positional arguments:
            * proto -- str, next layer protocol name
            * length -- int, valid (not padding) length

        Returns:
            * bool -- flag if extraction of next layer succeeded
            * Info -- info of next layer
            * ProtoChain -- protocol chain of next layer
            * str -- alias of next layer

        Protocols:
            * ARP -- data link layer
            * RARP -- data link layer
            * VLAN -- data link layer
            * IPv4 -- internet layer
            * IPv6 -- internet layer
            * IPX -- internet layer

        """
        if self._sigterm:
            from jspcap.protocols.raw import Raw as Protocol
        elif proto == 'ARP':
            from jspcap.protocols.link.arp import ARP as Protocol
        elif proto == 'RARP':
            from jspcap.protocols.link.rarp import RARP as Protocol
        elif proto == 'VLAN':
            from jspcap.protocols.link.vlan import VLAN as Protocol
        elif proto == 'IPv4':
            from jspcap.protocols.internet.ipv4 import IPv4 as Protocol
        elif proto == 'IPv6':
            from jspcap.protocols.internet.ipv6 import IPv6 as Protocol
        elif proto == 'IPX':
            from jspcap.protocols.internet.ipx import IPX as Protocol
        else:
            from jspcap.protocols.raw import Raw as Protocol
        next_ = Protocol(io.BytesIO(self._read_fileng(length)), length,
                            error=self._onerror, layer=self._exlayer, protocol=self._exproto)
        return True, next_.info, next_.protochain, next_.alias
